"""Processes a CRD file.

Note: Interfacing with external files is done in the `interfacer.py` library.
"""

import datetime
from pathlib import Path
import sys
from typing import Any, List, Tuple, Union
import warnings

import numpy as np

from . import processor_utils
from .data_io.crd_reader import CRDReader
from .utilities import ini, peirce, utils


class CRDFileProcessor:
    """Process a CRD file in this class, dead time corrections, etc.

    Computationally expensive routines are sourced out into processor_utils.py for
    jitting.

    Example:
        >>> my_file = Path("crd_files/my_file.crd")
        >>> crd = CRDFileProcessor(my_file)  # open the file
        >>> crd.spectrum_full()  # create a spectrum
    """

    def __init__(self, fname: Path) -> None:
        """Initialize the processor and read the CRD file that is wanted.

        :param fname: Filename of CRD file to be processed
        :type fname: Path
        """
        # read in the CRD file
        self.fname = fname
        self.crd = CRDReader(fname)
        self.ions_per_shot = self.crd.ions_per_shot
        self.ions_to_tof_map = self.crd.ions_to_tof_map
        self.all_tofs = self.crd.all_tofs

        # Data, ToF, and Masses
        self.tof = None
        self.mass = None
        self.data = None
        self.data_pkg = None

        # dictionary for what was run already - to be saved out
        self.applied_filters = {}

        # variables for filtered packages
        self._filter_max_ion_per_pkg_applied = False  # was max ions per pkg run?
        self._pkg_size = None  # max ions filtered with
        self._filter_max_ion_per_pkg_ind = None  # indices of pkgs that were trashed

        # Integrals
        self.integrals = None
        self.integrals_delta = None
        self.integrals_pkg = None
        self.integrals_delta_pkg = None

        # parameters for calibration and evaluation
        self._params_mcal = None  # mass calibration
        self._params_integrals = None  # integral definitions
        self._params_backgrounds = None  # bg_correction
        self._peak_fwhm = 0.0646  # peak fwhm in us
        self._us_to_chan = None  # how to change microseconds to channel / bin number

        # file info
        self.nof_shots = self.crd.nof_shots
        self.nof_shots_pkg = None

    # PROPERTIES #

    @property
    def def_backgrounds(self) -> Tuple[List[str], np.ndarray]:
        """Background definitions for integrals.

        The definitions consist of a tuple of a list and a np.ndarray.
        The list contains first the names of the integrals.
        The np.ndarray then contains in each row the lower and upper limit in amu of
        the peak that needs to be integrated.

        .. note:: The format for defining backgrounds is the same as the format for
            defining integrals, except that peaks can occur multiple times for
            multiple backgrounds.

        :return: Background definitions.

        :raise ValueError: Data Shape is wrong

        Example:
            >>> data = CRDFileProcessor("my_data.crd")
            >>> peak_names = ["54Fe", "54Fe"]
            >>> peak_limits = np.array([[53.4, 53.6], [54.4, 54.6]])
            >>> data.def_integrals = (peak_names, peak_limits)
        """
        return self._params_backgrounds

    @def_backgrounds.setter
    def def_backgrounds(self, value):
        if not value:  # empty list is passed
            self._params_backgrounds = None
        else:
            if len(value) != 2:
                raise ValueError("Data tuple must be of length 2.")
            if len(value[0]) != len(value[1]):
                raise ValueError("Name and data array must have the same length.")
            if value[1].shape[1] != 2:
                raise ValueError("The data array must have 2 entries for every line.")

            self._params_backgrounds = value

    @property
    def def_mcal(self) -> np.ndarray:
        """Mass calibration definitions.

        :return: Mass calibration definitions. The columns are as following:
            1st: ToF (us)
            2nd: Mass (amu)

        :raise TypeError: Value is not a numpy ndarray
        :raise ValueError: At least two parameters must be given for a mass calibration.
        :raise ValueError: The array is of the wrong shape.
        """
        return self._params_mcal

    @def_mcal.setter
    def def_mcal(self, value):
        if not isinstance(value, np.ndarray):
            raise TypeError(
                f"Mass calibration definition must be given as a numpy "
                f"ndarray but is a {type(value)}."
            )
        if value.shape[0] < 2:
            raise ValueError("At least two mass calibration points must be given.")
        if value.shape[1] != 2:
            raise ValueError("The mass calibration definition is of the wrong shape.")
        self._params_mcal = value

    @property
    def def_integrals(self) -> Tuple[List[str], np.ndarray]:
        """Integral definitions.

        The definitions consist of a tuple of a list and a np.ndarray.
        The list contains first the names of the integrals.
        The np.ndarray then contains in each row the lower and upper limit in amu of
        the peak that needs to be integrated.

        :return: Integral definitions.

        :raise ValueError: Data Shape is wrong
        :raise ValueError: More than one definition exist for a given peak.

        Example:
            >>> data = CRDFileProcessor("my_data.crd")
            >>> peak_names = ["54Fe", "64Ni"]
            >>> peak_limits = np.array([[53.8, 54.2], [63.5, 64.5]])
            >>> data.def_integrals = (peak_names, peak_limits)
        """
        return self._params_integrals

    @def_integrals.setter
    def def_integrals(self, value):
        if not value:  # empty list is passed
            self._params_integrals = None
            self._params_backgrounds = None
        else:
            if len(value) != 2:
                raise ValueError("Data tuple must be of length 2.")
            if len(value[0]) != len(value[1]):
                raise ValueError("Name and data array must have the same length.")
            if value[1].shape[1] != 2:
                raise ValueError("The data array must have 2 entries for every line.")
            if len(value[0]) != len(set(value[0])):
                raise ValueError(
                    "The peak names for integral definitions must be unique."
                )

            # delete backgrounds that are unused
            if self._params_backgrounds is not None:
                ind_to_delete = []
                for it, name in enumerate(self._params_backgrounds[0]):
                    if name not in value[0]:
                        ind_to_delete.append(it)
                if len(ind_to_delete) == len(self._params_backgrounds[0]):
                    self._params_backgrounds = None
                else:  # remove existing
                    bg_names = [
                        ele
                        for id, ele in enumerate(self._params_backgrounds[0])
                        if id not in ind_to_delete
                    ]
                    bg_vals = np.delete(
                        self._params_backgrounds[1], ind_to_delete, axis=0
                    )
                    self._params_backgrounds = bg_names, bg_vals

            self._params_integrals = value

    @property
    def peak_fwhm(self) -> float:
        """Get / Set the FWHM of the peak.

        :return: FWHM of the peak in us.
        """
        return self._peak_fwhm

    @peak_fwhm.setter
    def peak_fwhm(self, value: float) -> None:
        self._peak_fwhm = value

    @property
    def timestamp(self) -> datetime.datetime:
        """Get the time stamp when the recording was started.

        :return: Timestamp of the CRD file.

        Example:
            >>> crd = CRDFileProcessor(Path("my_file.crd"))
            >>> crd.timestamp
            datetime.datetime(2021, 7, 10, 11, 41, 13)
        """
        hdr_timestamp = self.crd.header["startDateTime"].rstrip(b"\x00").decode("utf-8")
        dt = datetime.datetime.strptime(hdr_timestamp, "%Y:%m:%d %H:%M:%S")
        return dt

    @property
    def us_to_chan(self) -> float:
        """Conversion factor for microseconds to channel / bin number.

        :return: Conversion factor
        """
        return self._us_to_chan

    @us_to_chan.setter
    def us_to_chan(self, value: float) -> None:
        self._us_to_chan = value

    # METHODS #

    def apply_individual_shots_filter(self, shots_rejected: np.ndarray):
        """Routine to finish filtering for individual shots.

        This will end up setting all the data. All routines that filter shots only
        have to provide a list of rejected shots. This routine does the rest, including.
        the handling of the data if packages exist.

        ToDo: rejected shots should be stored somewhere.

        :param shots_rejected: Indices of rejected shots.
        """
        len_indexes = len(self.ions_per_shot)

        # reject filtered packages, i.e., remove ions from deleted packages
        if self._filter_max_ion_per_pkg_applied:
            (
                shots_indexes,
                shots_rejected,
            ) = processor_utils.remove_shots_from_filtered_packages_ind(
                shots_rejected,
                len_indexes,
                self._filter_max_ion_per_pkg_ind,
                self._pkg_size,
            )
        else:
            shots_indexes = utils.not_index(shots_rejected, len_indexes)

        all_tofs_filtered = self._all_tofs_filtered(shots_indexes)

        self.data = processor_utils.sort_data_into_spectrum(
            all_tofs_filtered,
            self.crd.all_tofs.min(),
            self.crd.all_tofs.max(),
        )

        # remove the rejected shots from packages
        if self.data_pkg is not None:
            (
                self.data_pkg,
                self.nof_shots_pkg,
            ) = processor_utils.remove_shots_from_packages(
                self._pkg_size,
                shots_rejected,
                self.ions_to_tof_map,
                self.all_tofs,
                self.data_pkg,
                self.nof_shots_pkg,
                self._filter_max_ion_per_pkg_ind,
            )

        self.ions_per_shot = self.ions_per_shot[shots_indexes]
        self.ions_to_tof_map = self.ions_to_tof_map[shots_indexes]
        self.nof_shots = len(shots_indexes)

    def calculate_applied_filters(self):
        """Check for which filters are available and then recalculate all from start."""
        self.spectrum_full()  # reset all filters

        def get_arguments(key: str) -> Union[List, None]:
            """Get arguments from the dictionary or None.

            :param key: Key in dictionary ``self.applied_filters``

            :return: List if the key exists, None otherwise
            """
            try:
                return self.applied_filters[key]
            except KeyError:
                return

        # reset packages if not toggled
        if vals := get_arguments("packages"):
            if not vals[0]:
                self.data_pkg = None
                self._filter_max_ion_per_pkg_applied = False
                self._pkg_size = None
                self._filter_max_ion_per_pkg_ind = None
                self.integrals_pkg = None
                self.nof_shots_pkg = None

        # run through calculations
        if vals := get_arguments("spectrum_part"):
            if vals[0]:
                self.spectrum_part(vals[1])

        if vals := get_arguments("max_ions_per_shot"):
            if vals[0]:
                self.filter_max_ions_per_shot(vals[1])

        if vals := get_arguments("max_ions_per_time"):
            if vals[0]:
                self.filter_max_ions_per_time(vals[1], vals[2])

        if vals := get_arguments("max_ions_per_tof_window"):
            if vals[0]:
                self.filter_max_ions_per_tof_window(vals[1], np.array(vals[2]))

        if vals := get_arguments("packages"):
            if vals[0]:
                self.packages(vals[1])

        if vals := get_arguments("max_ions_per_pkg"):
            if vals[0]:
                self.filter_max_ions_per_pkg(vals[1])

        # fixme: after peirce criterion is done!
        # if get_arguments("pkg_peirce_rejection"):
        #     self.filter_pkg_peirce_countrate()

        if vals := get_arguments("macro"):
            if vals[0]:
                self.run_macro(Path(vals[1]))

        if vals := get_arguments("dead_time_corr"):
            if vals[0]:
                self.dead_time_correction(vals[1])

    def dead_time_correction(self, dbins: int) -> None:
        """Perform a dead time correction on the whole spectrum.

        If packages were set, the dead time correction is performed on each package
        individually as well.
        :param dbins: Number of dead bins after original bin (total - 1).

        :warning.warn: There are no shots left in the package. No deadtime
            correction can be applied.
        """
        self.applied_filters["dead_time_corr"] = [True, dbins]

        if self.nof_shots == 0:
            warnings.warn(
                "No data available; maybe all shots were filtered out?", UserWarning
            )
            return

        self.data = processor_utils.dead_time_correction(
            self.data.reshape(1, self.data.shape[0]),
            np.array(self.nof_shots).reshape(1),
            dbins,
        )[
            0
        ]  # want to shape it back the way it was!

        if self.data_pkg is not None:
            self.data_pkg = processor_utils.dead_time_correction(
                self.data_pkg, self.nof_shots_pkg, dbins
            )

    def filter_max_ions_per_pkg(self, max_ions: int) -> None:
        """Filter out packages with too many ions.

        .. note:: Only run more than once if filtering out more. Otherwise, you need
            to reset the dataset first.

        :param max_ions: Maximum number of ions per package.

        :raises ValueError: Invalid range for number of ions.
        :raises OSError: No package data available.
        """
        if max_ions < 1:
            raise ValueError("The maximum number of ions must be larger than 1.")
        if self.data_pkg is None:
            raise OSError("There is no packaged data. Please create packages first.")

        # update filter dictionary
        self.applied_filters["max_ions_per_pkg"] = [True, max_ions]

        # update helper variables
        self._filter_max_ion_per_pkg_applied = True

        total_ions_per_pkg = np.sum(self.data_pkg, axis=1)

        self._filter_max_ion_per_pkg_ind = np.where(total_ions_per_pkg > max_ions)[0]

        self.data_pkg = np.delete(
            self.data_pkg, self._filter_max_ion_per_pkg_ind, axis=0
        )
        self.nof_shots_pkg = np.delete(
            self.nof_shots_pkg, self._filter_max_ion_per_pkg_ind, axis=0
        )

        self.data = np.sum(self.data_pkg, axis=0)
        self.nof_shots = np.sum(self.nof_shots_pkg)

    def filter_max_ions_per_shot(self, max_ions: int) -> None:
        """Filter out shots that have more than the max_ions defined.

        .. note:: Only run more than once if filtering out more. Otherwise, you need
            to reset the dataset first.

        :param max_ions: Maximum number of ions allowed in a shot.

        :raises ValueError: Invalid range for number of ions.
        """
        if max_ions < 1:
            raise ValueError("The maximum number of ions must be >=1.")

        self.applied_filters["max_ions_per_shot"] = [True, max_ions]

        shots_rejected = np.where(self.ions_per_shot > max_ions)[0]

        self.apply_individual_shots_filter(shots_rejected)

    def filter_max_ions_per_time(self, max_ions: int, time_us: float) -> None:
        """Filter shots with >= max ions per time, i.e., due to ringing.

        :param max_ions: Maximum number of ions that is allowed within a time window.
        :param time_us: Width of the time window in microseconds (us)
        """
        self.applied_filters["max_ions_per_time"] = [True, max_ions, time_us]

        time_chan = int(time_us * self.us_to_chan)

        shots_to_check = np.where(self.ions_per_shot > max_ions)[0]

        if shots_to_check.shape == (0,):  # nothing needs to be done
            return

        all_tofs_filtered = self._all_tofs_filtered(shots_to_check)

        shot_mask = processor_utils.mask_filter_max_ions_per_time(
            self.ions_per_shot[shots_to_check], all_tofs_filtered, max_ions, time_chan
        )
        shots_rejected = shots_to_check[shot_mask]

        if shots_rejected.shape != (0,):
            self.apply_individual_shots_filter(shots_rejected)

    def filter_max_ions_per_tof_window(
        self, max_ions: int, tof_window: np.ndarray
    ) -> None:
        """Filer out maximum number of ions in a given ToF time window.

        :param max_ions: Maximum number of ions in the time window.
        :param tof_window: The time of flight window that the ions would have to be in.
            Array of start and stop time of flight (2 entries).

        :raises ValueError: Length of `tof_window` is wrong.
        """
        if len(tof_window) != 2:
            raise ValueError(
                "ToF window must be specified with two entries: the start "
                "and the stop time of the window."
            )

        if not isinstance(tof_window, np.ndarray):
            tof_window = np.array(tof_window)

        self.applied_filters["max_ions_per_tof_window"] = [
            True,
            max_ions,
            tof_window.tolist(),
        ]

        # convert to int to avoid weird float issues
        channel_window = np.array(tof_window * self.us_to_chan, dtype=int)

        shots_to_check = np.where(self.ions_per_shot > max_ions)[0]

        if shots_to_check.shape == (0,):  # nothing needs to be done
            return

        all_tofs_filtered = self._all_tofs_filtered(shots_to_check)

        shot_mask = processor_utils.mask_filter_max_ions_per_tof_window(
            self.ions_per_shot[shots_to_check],
            all_tofs_filtered,
            max_ions,
            channel_window,
        )
        shots_rejected = shots_to_check[shot_mask]

        if shots_rejected.shape != (0,):
            self.apply_individual_shots_filter(shots_rejected)

    def filter_pkg_peirce_countrate(self) -> None:
        """Filter out packages based on Peirce criterion for total count rate.

        Fixme: This needs more thinking and testing
        Now we are going to directly use all the integrals to get the sum of the counts,
        which we will then feed to the rejection routine. Maybe this can detect blasts.

        .. warning:: Running this more than once might lead to weird results. You have
            been warned!

        """  # noqa: D202

        warnings.warn(
            "This routine to reject packages according to the Peirce criterium is "
            "largely untested.",
            UserWarning,
        )

        self.applied_filters["pkg_peirce_rejection"] = True

        sum_integrals = self.integrals_pkg[:, :, 0].sum(axis=1)
        _, _, _, rejected_indexes = peirce.reject_outliers(sum_integrals)

        print(
            f"Peirce criterion rejected "
            f"{len(rejected_indexes)} / {len(self.integrals_pkg)} "
            f"packages"
        )
        index_list = list(map(int, rejected_indexes))
        integrals_pkg = np.delete(self.integrals_pkg, index_list, axis=0)
        self.nof_shots_pkg = np.delete(self.nof_shots_pkg, index_list)
        self.nof_shots = np.sum(self.nof_shots_pkg)

        # integrals
        integrals = np.zeros_like(self.integrals)
        integrals[:, 0] = integrals_pkg.sum(axis=0)[:, 0]
        integrals[:, 1] = np.sqrt(np.sum(integrals_pkg[:, :, 1] ** 2, axis=0))

        # write back
        self.integrals = integrals
        self.integrals_pkg = integrals_pkg

    def integrals_calc(self, bg_corr=True) -> None:
        """Calculate integrals for data and packages (if present).

        The integrals to be set per peak are going to be set as an ndarray.
        Each row will contain one entry in the first column and its associated
        uncertainty in the second.

        :param bg_corr: If false, will never do background correction. Otherwise
            (default), background correction will be applied if available. This is a
            toggle to switch usage while leaving backgrounds defined.

        :raises ValueError: No integrals were set.
        :raises ValueError: No mass calibration has been applied.
        """

        def integral_windows(limits_tmp: np.array) -> List:
            """Create windows list for given limits.

            :param limits_tmp: Window limits.

            :return: List with all the windows that need to be calculated.
            """
            windows_tmp = []
            for low_lim, upp_lim in limits_tmp:
                windows_tmp.append(
                    np.where(np.logical_and(self.mass >= low_lim, self.mass <= upp_lim))
                )
            return windows_tmp

        if self._params_integrals is None:
            raise ValueError("No integrals were set.")
        if self.mass is None:
            raise ValueError("A mass calibration needs to be applied first.")

        names, limits = self.def_integrals

        windows = integral_windows(limits)

        self.integrals, self.integrals_pkg = processor_utils.integrals_summing(
            self.data, tuple(windows), self.data_pkg
        )

        # background correction
        if bg_corr and self._params_backgrounds is not None:
            names_bg, limits_bg = self.def_backgrounds

            windows_bgs = integral_windows(limits_bg)

            bgs, bgs_pkg = processor_utils.integrals_summing(
                self.data, tuple(windows_bgs), self.data_pkg
            )

            # determine channel lengths
            peak_ch_length = np.array([len(it) for it in windows])
            bgs_ch_length = np.array([len(it) for it in windows_bgs])

            # call the processor and do the background correction
            self.integrals, self.integrals_pkg = processor_utils.integrals_bg_corr(
                self.integrals,
                np.array(names),
                peak_ch_length,
                bgs,
                np.array(names_bg),
                bgs_ch_length,
                self.integrals_pkg,
                bgs_pkg,
            )

    def integrals_calc_delta(self) -> None:
        """Calculate delta values for integrals and save them in class.

        This routine uses the ``iniabu`` package to calculate delta values for defined
        integrals. It reads the peak names and calculates delta values for isotopes
        that can be understood ``iniabu``, and calculates the delta values with
        respect to the major isotope. These values are then saved to the class as
        ``integrals_delta`` and ``integrals_delta_pkg``, if packages were defined.
        Uncertainties are propagated according to Gaussian error propagation.
        The format of the resulting arrays are identical to the ``integrals`` and
        ``integrals_pkg`` arrays.

        :raises ValueError: No integrals were calculated.
        """
        if self.integrals is None or self.def_integrals is None:
            raise ValueError("No integrals were defined or calculated.")

        peak_names = self.def_integrals[0]

        integrals_delta = processor_utils.delta_calc(peak_names, self.integrals)

        if self.integrals_pkg is not None:
            integrals_delta_pkg = np.zeros_like(self.integrals_pkg, dtype=float)
            for it, line in enumerate(self.integrals_pkg):
                integrals_delta_pkg[it] = processor_utils.delta_calc(peak_names, line)
            self.integrals_delta_pkg = integrals_delta_pkg

        self.integrals_delta = integrals_delta

    def mass_calibration(self) -> None:
        r"""Perform a mass calibration on the data.

        Let m be the mass and t the respective time of flight. We can then write:

            .. math::
                t \propto \sqrt[a]{m}

        Usually it is assumed that $a=2$, i.e., that the square root is taken.
        We don't have to assume this though. In the generalized form we can now
        linearize the mass calibration such that:

            .. math::
                \log(m) = a \log(t) + b

        Here, :math:`a` is, as above, the exponent, and :math:`b` is a second constant.
        With two values or more for :math:`m` and :math:`t`, we can then make a
        linear approximation for the mass calibration :math:`m(t)`.

        :raises ValueError: No mass calibration set.
        """
        if self._params_mcal is None:
            raise ValueError("No mass calibration was set.")

        self.mass = processor_utils.mass_calibration(self.def_mcal, self.tof)

    def optimize_mcal(self, offset: float = None) -> None:
        """Take an existing mass calibration and finds maxima within a FWHM.

        This will act on small corrections for drifts in peaks.

        :param offset: How far do you think the peak has wandered? If None, it will be
            set to the FWHM value.
        """
        if offset is None:
            offset = self.peak_fwhm

        positions = self.def_mcal[:, 0]
        positions_new = np.zeros_like(positions) * np.nan  # nan array

        for it, pos in enumerate(positions):
            min_time = pos - offset - 2 * self.peak_fwhm
            max_time = pos + offset + 2 * self.peak_fwhm
            if max_time > self.tof.max():  # we don't have a value here
                continue
            window = np.where(np.logical_and(self.tof > min_time, self.tof < max_time))
            tofs = self.tof[window]
            data = self.data[window]
            positions_new[it] = processor_utils.gaussian_fit_get_max(tofs, data)

        mcal_new = self.def_mcal.copy()
        index_to_del = []
        for it, posn in enumerate(positions_new):
            if np.abs(mcal_new[it][0] - posn) < offset:
                mcal_new[it][0] = posn
            else:
                index_to_del.append(it)

        mcal_new = np.delete(mcal_new, index_to_del, axis=0)
        if len(mcal_new) < 2:
            warnings.warn(
                "Automatic mass calibration optimization did not find enough peaks."
            )
        else:
            self.def_mcal = mcal_new

    def packages(self, shots: int) -> None:
        """Break data into packages.

        :param shots: Number of shots per package. The last package will have the rest.

        :raises ValueError: Number of shots out of range
        """
        if shots < 1 or shots >= self.nof_shots:
            raise ValueError(
                f"Number of shots per package must be between 1 and "
                f"{self.nof_shots}, but is {shots}."
            )

        self.applied_filters["packages"] = [True, shots]

        self._pkg_size = shots

        self.data_pkg, self.nof_shots_pkg = processor_utils.create_packages(
            shots, self.ions_to_tof_map, self.all_tofs
        )

    def run_macro(self, fname: Path) -> None:
        """Run your own macro.

        The macro will be imported here and then run. Details on how to write a macro
        can be found in the documentation.

        :param fname: Filename to the macro.
        """
        self.applied_filters["macro"] = [True, str(fname.absolute())]

        pyfile = fname.with_suffix("").name
        file_path = fname.absolute().parent
        print("fname")
        print(fname)

        sys.path.append(str(file_path))

        exec(f"import {pyfile}") in globals(), locals()  # noqa: S102
        macro = vars()[pyfile]
        macro.calc(self)

        sys.path.remove(str(file_path))

    def sort_integrals(self, sort_vals: bool = True) -> None:
        """Sort all the integrals that are defined by mass.

        Takes the integrals and the names and sorts them by proton number (first order),
        then by mass (second order). All integrals that cannot be identified with a
        clear proton number (e.g., molecules) are sorted in at the end of the primary
        sorting, then sorted by mass.
        The starting mass of each integral is used for sorting.
        If no integrals are defined, this routine does nothing.

        :param sort_vals: Sort the integrals and integral packages? Default: True

        Example:
            >>> crd.def_integrals
            ["Fe-56", "Ti-46"], array([[55.8, 56.2], [45.8, 46.2]])
            >>> crd.sort_integrals()
            >>> crd.def_integrals
            ["Ti-46", "Fe-56"], array([[45.8, 46.2], [55.8, 56.2]])
        """
        if self.def_integrals is None:
            return

        names, values = self.def_integrals

        zz = []  # number of protons per isotope - first sort key
        for name in names:
            try:
                zz.append(ini.iso[name].z)
            except IndexError:
                zz.append(999)  # at the end of everything

        sort_ind = sorted(np.arange(len(names)), key=lambda x: (zz[x], values[x, 0]))

        if (sort_ind == np.arange(len(names))).all():  # already sorted
            return

        names_sorted = list(np.array(names)[sort_ind])
        self.def_integrals = names_sorted, values[sort_ind]

        if self.integrals is not None and sort_vals:
            self.integrals = self.integrals[sort_ind]
        if self.integrals_pkg is not None and sort_vals:
            self.integrals_pkg = self.integrals_pkg[:, sort_ind]

    def spectrum_full(self) -> None:
        """Create ToF and summed ion count array for the full spectrum.

        The full spectrum is transfered to ToF and ion counts. The spectrum is then
        saved to:
        - ToF array is written to `self.tof`
        - Data array is written to `self.data`

        :warnings: Time of Flight and data have different shape
        """
        bin_length = self.crd.header["binLength"]
        bin_start = self.crd.header["binStart"]
        bin_end = self.crd.header["binEnd"]
        delta_t = self.crd.header["deltaT"]

        # reset the data
        self.ions_per_shot = self.crd.ions_per_shot
        self.ions_to_tof_map = self.crd.ions_to_tof_map
        self.all_tofs = self.crd.all_tofs
        self.nof_shots = self.crd.nof_shots

        # set up ToF
        self.tof = (
            np.arange(bin_start, bin_end + 1, 1) * bin_length / 1e6 + delta_t * 1e6
        )
        self.data = processor_utils.sort_data_into_spectrum(
            self.all_tofs, self.all_tofs.min(), self.all_tofs.max()
        )

        # set constants
        self.us_to_chan = 1e6 / self.crd.header["binLength"]  # convert us to bins

        if self.tof.shape != self.data.shape:
            warnings.warn(
                "Bin ranges in CRD file were of bad length. Creating ToF "
                "array without CRD header input."
            )
            self.tof = (
                np.arange(len(self.data)) * bin_length / 1e6
                + self.all_tofs.min() / self.us_to_chan
            )

    def spectrum_part(self, rng: Union[Tuple[Any], List[Any]]) -> None:
        """Create ToF for a part of the spectra.

        Select part of the shot range. These ranges will be 1 indexed! Always start
        with the full data range.

        :param rng: Shot range, either as a tuple (from, to) or as a tuple of multiple
            ((from1, to1), (from2, to2), ...).

        :raises ValueError: Ranges are not defined from, to where from < to
        :raises ValueError: Tuples are not mutually exclusive.
        :raises IndexError: One or more indexes are out of range.
        """
        self.applied_filters["spectrum_part"] = [True, rng]

        # reset current settings
        self.ions_to_tof_map = self.crd.ions_to_tof_map
        self.all_tofs = self.crd.all_tofs

        # range
        rng = np.array(rng)
        if len(rng.shape) == 1:  # only one entry
            rng = rng.reshape(1, 2)

        # check for index errors
        index_err_lower = False
        index_err_upper = False
        if (rng < 1).any():
            index_err_lower = True
        if (rng > int(self.crd.header["nofShots"])).any():
            index_err_upper = True

        index_err_both = False
        if index_err_upper and index_err_lower:
            index_err_both = True

        if index_err_lower or index_err_upper:
            msg = (
                f"Your{' lower' if index_err_lower else ''}"
                f"{' and' if index_err_both else ''}"
                f"{' upper' if index_err_upper else ''}"
                f"{' indexes are' if index_err_both else ' index is'} out of range.\n\n"
                f"The first shot (index 1) and the last index you select will be "
                f"included. For example, selecting '1,5' will include shots 1 to 5."
            )
            raise IndexError(msg)

        # subtract 1 from start range -> zero indexing plus upper limit inclusive now
        rng[:, 0] -= 1

        # sort by first entry
        rng = rng[rng[:, 0].argsort()]

        # check if any issues with the
        if any(rng[:, 1] < rng[:, 0]):
            raise ValueError(
                "The `from, to` values in your range are not defined "
                "such that `from` < `to`."
            )

        # check that mutually exclusive
        for it in range(1, len(rng)):
            if rng[it - 1][1] > rng[it][0]:
                raise ValueError("Your ranges are not mutually exclusive.")

        # filter ions per shot
        ion_indexes = processor_utils.multi_range_indexes(rng)

        # create all_tof ranges and filter
        rng_all_tofs = self.ions_to_tof_map[ion_indexes]

        tof_indexes = processor_utils.multi_range_indexes(rng_all_tofs)

        all_tofs_filtered = self.all_tofs[tof_indexes]
        ions_to_tof_map_filtered = self.ions_to_tof_map[ion_indexes]

        # if empty shape: we got no data!
        if len(tof_indexes) == 0:
            self.data = np.zeros_like(self.data)
        else:
            self.data = processor_utils.sort_data_into_spectrum(
                all_tofs_filtered, self.crd.all_tofs.min(), self.crd.all_tofs.max()
            )

        # set back values
        self.ions_per_shot = self.ions_per_shot[ion_indexes]
        self.ions_to_tof_map = ions_to_tof_map_filtered
        self.all_tofs = all_tofs_filtered
        self.nof_shots = len(ion_indexes)

    # PRIVATE ROUTINES #

    def _all_tofs_filtered(self, shots_indexes: np.array) -> np.array:
        """Filter time of flights based on the indexes of the shots.

        This function is heavily used in filters.

        :param shots_indexes: Array with indexes of the shots.

        :return: All time of flight bins for the given shots
        """
        rng_all_tofs = self.ions_to_tof_map[shots_indexes]
        tof_indexes = processor_utils.multi_range_indexes(rng_all_tofs)
        return self.all_tofs[tof_indexes]



"""Tests for processors utilities."""

import pytest

import matplotlib.pyplot as plt
import numpy as np

import rimseval.processor_utils as pu
import rimseval.data_io.crd_utils as crdu


def test_create_packages():
    """Create packages from data."""
    ions_per_shot = np.array([0, 0, 1, 0, 2, 0, 3, 2, 1, 4])
    all_tofs = np.array([20, 25, 70, 53, 68, 11, 54, 12, 68, 99, 65, 48, 7])
    bin_start = all_tofs.min()
    bin_end = all_tofs.max()
    len_data = bin_end - bin_start + 1
    tofs_mapper = crdu.shot_to_tof_mapper(ions_per_shot)
    assert ions_per_shot.sum() == len(all_tofs)
    assert all_tofs.max() < len_data + bin_start  # sanity checks for setup

    # packages expected
    pkg_length = 4
    nof_pkg = len(ions_per_shot) // pkg_length
    if len(ions_per_shot) % pkg_length > 0:
        nof_pkg += 1

    pkg_nof_shots_exp = np.zeros(nof_pkg) + pkg_length
    if (tmp := len(ions_per_shot) % pkg_length) > 0:
        pkg_nof_shots_exp[-1] = tmp

    pkg_data_exp = np.zeros((nof_pkg, len_data))

    for it, shot in enumerate(ions_per_shot):
        pkg_it = it // pkg_length
        mapper = tofs_mapper[it]
        tofs = all_tofs[mapper[0] : mapper[1]]
        for tof in tofs:
            pkg_data_exp[pkg_it][int(tof) - bin_start] += 1

    pkg_data_rec, pkg_nof_shots_rec = pu.create_packages(
        pkg_length, tofs_mapper, all_tofs
    )
    np.testing.assert_equal(pkg_nof_shots_rec, pkg_nof_shots_exp)
    np.testing.assert_equal(pkg_data_rec, pkg_data_exp)


def test_delta_calc():
    """Take an integrals like array and calculate delta values.

    Details of delta calculation is not tested.
    """
    names = ["Fe54", "Fe56", "244Pu", "bg"]
    integrals = np.array([[10000, 100], [100000, 240], [100, 10], [2001, 21]])
    deltas = pu.delta_calc(names, integrals)
    assert np.isnan(deltas[2:3]).all()  # last two must be nans


def test_gaussian_fit():
    """Do a Gaussian feed on fixed-seed random data."""
    np.random.seed(42)
    mu, sigma = 0, 0.1
    sampl = np.random.normal(mu, sigma, 1000)

    ydat, bin_edges, _ = plt.hist(sampl, 25)
    xdat = (bin_edges[1:] + bin_edges[:-1]) / 2

    max_calc = pu.gaussian_fit_get_max(xdat, ydat)
    assert mu == pytest.approx(max_calc, abs=0.01)


def test_integrals_bg_corr():
    """Background correction for defined integrals."""
    integrals = np.array([[10, np.sqrt(10)], [40, np.sqrt(40)]])
    int_names = np.array(["28Si", "29Si"])
    int_ch = np.array([30, 40])
    bgs = np.array([[1, np.sqrt(1)], [2, np.sqrt(2)], [3, np.sqrt(3)]])
    bgs_names = np.array(["28Si", "28Si", "29Si"])
    bgs_ch = np.array([20, 10, 50])

    # manual calculations for expected values
    bgs_cnt = bgs[:, 0]
    bgs_norm = np.array(
        [(bgs_cnt[0] / bgs_ch[0] + bgs_cnt[1] / bgs_ch[1]) / 2, bgs_cnt[2] / bgs_ch[2]]
    )
    bgs_norm_unc = np.array(
        [
            (np.sqrt(bgs_cnt[0]) / bgs_ch[0] + np.sqrt(bgs_cnt[1]) / bgs_ch[1]) / 2,
            np.sqrt(bgs_cnt[2]) / bgs_ch[2],
        ]
    )
    integrals_corr = integrals[:, 0] - int_ch * bgs_norm
    integrals_corr_unc = np.sqrt(integrals[:, 0] + bgs_norm_unc**2)

    integrals_exp = np.array(
        [
            [integrals_corr[it], integrals_corr_unc[it]]
            for it in range(len(integrals_corr))
        ]
    )

    integrals_rec, integrals_pkg_rec = pu.integrals_bg_corr(
        integrals, int_names, int_ch, bgs, bgs_names, bgs_ch
    )
    np.testing.assert_equal(integrals_rec, integrals_exp)
    assert integrals_pkg_rec is None


def test_integrals_bg_corr_pkg():
    """Background correction for defined integrals with packages."""
    integrals = np.array([[10, np.sqrt(10)], [40, np.sqrt(40)]])
    integrals_pkg = np.array([integrals])
    int_names = np.array(["28Si", "29Si"])
    int_ch = np.array([30, 40])
    bgs = np.array([[1, np.sqrt(1)], [2, np.sqrt(2)], [3, np.sqrt(3)]])
    bgs_pkg = np.array([bgs])
    bgs_names = np.array(["28Si", "28Si", "29Si"])
    bgs_ch = np.array([20, 10, 50])

    # manual calculations for expected values
    bgs_cnt = bgs[:, 0]
    bgs_norm = np.array(
        [(bgs_cnt[0] / bgs_ch[0] + bgs_cnt[1] / bgs_ch[1]) / 2, bgs_cnt[2] / bgs_ch[2]]
    )
    bgs_norm_unc = np.array(
        [
            (np.sqrt(bgs_cnt[0]) / bgs_ch[0] + np.sqrt(bgs_cnt[1]) / bgs_ch[1]) / 2,
            np.sqrt(bgs_cnt[2]) / bgs_ch[2],
        ]
    )
    integrals_corr = integrals[:, 0] - int_ch * bgs_norm
    integrals_corr_unc = np.sqrt(integrals[:, 0] + bgs_norm_unc**2)

    integrals_exp = np.array(
        [
            [integrals_corr[it], integrals_corr_unc[it]]
            for it in range(len(integrals_corr))
        ]
    )

    # no packages
    integrals_rec, integrals_pkg_rec = pu.integrals_bg_corr(
        integrals, int_names, int_ch, bgs, bgs_names, bgs_ch
    )
    np.testing.assert_equal(integrals_rec, integrals_exp)
    assert integrals_pkg_rec is None

    # packages
    integrals_rec, integrals_pkg_rec = pu.integrals_bg_corr(
        integrals, int_names, int_ch, bgs, bgs_names, bgs_ch, integrals_pkg, bgs_pkg
    )
    np.testing.assert_equal(integrals_rec, integrals_exp)
    np.testing.assert_equal(integrals_pkg_rec[0], integrals_exp)
    assert len(integrals_pkg_rec) == 1


def test_integrals_summing():
    """Sum integrals for given data."""
    data = np.arange(100) * 2  # *2 such that index not equal to value
    window = (np.arange(10), np.arange(5) + 50)  # two windows

    integral_values = [np.sum(data[0:10]), np.sum(data[50:55])]
    integrals_exp = np.array([[it, np.sqrt(it)] for it in integral_values])

    integrals_rec, integrals_pkg_rec = pu.integrals_summing(data, window)
    np.testing.assert_equal(integrals_rec, integrals_exp)
    assert integrals_pkg_rec is None


def test_integrals_summing_pkg():
    """Sum integrals for given data and packages."""
    data = np.arange(100) * 2  # *2 such that index not equal to value
    data_pkg = np.array([data])
    window = (np.arange(10), np.arange(5) + 50)  # two windows

    integral_values = [np.sum(data[0:10]), np.sum(data[50:55])]
    integrals_exp = np.array([[it, np.sqrt(it)] for it in integral_values])

    integrals_rec, integrals_pkg_rec = pu.integrals_summing(data, window, data_pkg)
    np.testing.assert_equal(integrals_rec, integrals_exp)
    np.testing.assert_equal(integrals_pkg_rec[0], integrals_exp)


def test_mask_filter_max_ions_per_time():
    """Filter maximum number of ions per time window."""
    ions_per_shot = np.array([4, 0, 4, 5, 4])
    tofs = np.array(
        [1, 2, 3, 4] + [] + [1, 3, 5, 10] + [10, 15, 20, 25, 30] + [9, 10, 11, 15]
    )  # looks weird, but easier to stich together by hand
    max_ions = 3
    time_window_chan = 4

    exp_mask = np.array([0, 4])  # where conditions are met
    rec_mask = pu.mask_filter_max_ions_per_time(
        ions_per_shot, tofs, max_ions, time_window_chan
    )
    np.testing.assert_equal(exp_mask, rec_mask)


def test_mask_filter_max_ions_per_tof_window():
    """Filter maximum number of ions per time window."""
    ions_per_shot = np.array([4, 0, 4, 5, 4])
    tofs = np.array(
        [1, 2, 3, 4] + [] + [1, 3, 5, 10] + [10, 15, 20, 25, 30] + [9, 10, 11, 15]
    )  # looks weird, but easier to stich together by hand
    max_ions = 2
    time_window_chan = np.array([1, 5])

    exp_mask = np.array([0, 2])  # where conditions are met
    rec_mask = pu.mask_filter_max_ions_per_tof_window(
        ions_per_shot, tofs, max_ions, time_window_chan
    )
    np.testing.assert_equal(exp_mask, rec_mask)


def test_mass_calibration():
    """Perform a mass calibration, functionality tested in unit tests."""
    params = np.array([[1, 1], [10, 10], [20, 28.35]])
    tofs = np.arange(0, 20)

    mass, params_fit = pu.mass_calibration(params, tofs, return_params=True)

    assert tofs.shape == mass.shape
    assert params_fit is not None


def test_mass_to_tof():
    """Convert mass to time of flight."""
    m = 42.0
    tm0 = 13.0
    const = 0.7

    result_exp = np.sqrt(m) * const + tm0

    assert result_exp == pu.mass_to_tof(m, tm0, const)


def test_multi_range_indexes():
    """Create multi-range indexes for view on data."""
    ranges = np.array([[0, 2], [3, 6], [0, 0], [15, 22]])
    indexes_exp = np.array([0, 1, 3, 4, 5, 15, 16, 17, 18, 19, 20, 21], dtype=int)

    indexes_rec = pu.multi_range_indexes(ranges)
    np.testing.assert_equal(indexes_rec, indexes_exp)
    assert indexes_rec.dtype == int


def test_multi_range_indexes_empty():
    """Return empty index array if all ranges of zero length."""
    ranges = np.array([[0, 0], [0, 0]])
    indexes_exp = np.array([])
    np.testing.assert_equal(pu.multi_range_indexes(ranges), indexes_exp)


def test_sort_data_into_spectrum():
    """Sort some small data into a spectrum."""
    ions = np.array([0, 1, 5, 10, 9])
    assert ions.min() == 0  # requirement for this simple test
    spectrum_exp = np.zeros(ions.max() - ions.min() + 1)
    for ion in ions:
        spectrum_exp[ion] += 1
    spectrum_rec = pu.sort_data_into_spectrum(ions, ions.min(), ions.max())
    np.testing.assert_equal(spectrum_rec, spectrum_exp)



"""Unit tests for processor, focusing on multiple functions at a time."""

from pathlib import Path

import numpy as np
import pytest

import rimseval
from rimseval import interfacer
from rimseval.processor import CRDFileProcessor


def test_filter_pkg_peirce_countrate(crd_file):
    """Apply Peirce countrate outlier rejection (run through, no assertions)."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(2)

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (["all"], np.array([[0.9, 2.1]]))  # avoid floating errors
    crd.integrals_calc()

    crd.filter_pkg_peirce_countrate()


def test_integrals(crd_file):
    """Define an integral manually and calculate the integration."""
    _, ions_per_shot, all_tofs, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (["all"], np.array([[0.9, 2.1]]))  # avoid floating errors
    crd.integrals_calc()

    assert len(all_tofs) == crd.integrals[0][0]
    assert np.sqrt(len(all_tofs)) == pytest.approx(crd.integrals[0][1])


def test_integrals_bg_corr_behavior(crd_file):
    """Ensure that bg corrected integrals behave correctly."""
    _, ions_per_shot, all_tofs, fname = crd_file

    shots_per_pkg = 2
    nof_pkgs = int(np.ceil(len(ions_per_shot) / shots_per_pkg))
    integrals_exp = np.zeros((nof_pkgs, 1, 2))  # 1 integral

    start_ind = 0
    for it in range(nof_pkgs - 1):
        stop_ind = start_ind + shots_per_pkg
        integrals_exp[it][0][0] = np.sum(ions_per_shot[start_ind:stop_ind])
        integrals_exp[it][0][1] = np.sqrt(integrals_exp[it][0][0])
        start_ind = stop_ind
    # add the last
    integrals_exp[-1][0][0] = np.sum(ions_per_shot[start_ind:])
    integrals_exp[-1][0][1] = np.sqrt(integrals_exp[-1][0][0])

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(shots_per_pkg)

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (["all"], np.array([[0.9, 2.1]]))  # avoid floating errors
    crd.integrals_calc()

    integrals_only = np.array(crd.integrals)
    integrals_pkg_only = np.array(crd.integrals_pkg)

    # set the background correction
    crd.def_backgrounds = (["all"], np.array([[0.1, 0.9]]))
    crd.integrals_calc()

    # now make sure that integrals are always smaller when bg corrected than when not
    assert all(crd.integrals[:, 0] <= integrals_only[:, 0])
    assert all(crd.integrals[:, 1] >= integrals_only[:, 1])

    # sum of packaged integrals is still equal to sum of integrals
    np.testing.assert_allclose(crd.integrals_pkg.sum(axis=0)[:, 0], crd.integrals[:, 0])


def test_integrals_multiple_backgrounds(crd_file):
    """Run bg correction with multiple backgrounds - run through only."""
    _, ions_per_shot, all_tofs, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (["all"], np.array([[0.9, 2.1]]))  # avoid floating errors
    crd.integrals_calc()

    # set the background correction
    crd.def_backgrounds = (["all", "all"], np.array([[0.1, 0.9], [2.1, 3.0]]))
    crd.integrals_calc(bg_corr=True)


def test_background_one_bg_with_multiple_peaks(crd_data):
    """Ti Standard 01 spectrum with json file to do background correction."""
    crd = CRDFileProcessor(crd_data.joinpath("ti_standard_01.crd"))
    interfacer.load_cal_file(crd)
    crd.spectrum_full()
    crd.mass_calibration()
    crd.calculate_applied_filters()
    crd.integrals_calc(bg_corr=True)

    # assert background correction does not throw nans or zeros
    assert not np.isnan(crd.integrals).any()
    assert not (crd.integrals == 0).any()


def test_integrals_pkg(crd_file):
    """Define an integral manually and calculate the integration for packages."""
    _, ions_per_shot, all_tofs, fname = crd_file

    shots_per_pkg = 2
    nof_pkgs = int(np.ceil(len(ions_per_shot) / shots_per_pkg))
    integrals_exp = np.zeros((nof_pkgs, 1, 2))  # 1 integral

    start_ind = 0
    for it in range(nof_pkgs - 1):
        stop_ind = start_ind + shots_per_pkg
        integrals_exp[it][0][0] = np.sum(ions_per_shot[start_ind:stop_ind])
        integrals_exp[it][0][1] = np.sqrt(integrals_exp[it][0][0])
        start_ind = stop_ind
    # add the last
    integrals_exp[-1][0][0] = np.sum(ions_per_shot[start_ind:])
    integrals_exp[-1][0][1] = np.sqrt(integrals_exp[-1][0][0])

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(shots_per_pkg)

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (["all"], np.array([[0.9, 2.1]]))  # avoid floating errors
    crd.integrals_calc()

    np.testing.assert_almost_equal(crd.integrals_pkg, integrals_exp)

    # check that sum agrees -> sqrt of sqsum for uncertainty
    crd_integrals_sum = np.array(crd.integrals_pkg)
    crd_integrals_sum[:, :, 1] = crd_integrals_sum[:, :, 1] ** 2
    crd_integrals_sum = crd_integrals_sum.sum(axis=0)
    crd_integrals_sum[:, 1] = np.sqrt(crd_integrals_sum[:, 1])

    np.testing.assert_almost_equal(crd.integrals, crd_integrals_sum)


def test_integrals_pkg_with_filtering(crd_file):
    """Filtering in packages and get the sum of the integrals."""
    _, _, _, fname = crd_file
    shots_per_pkg = 1
    max_ions_per_shot = 1

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(shots_per_pkg)
    crd.filter_max_ions_per_pkg(1)

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (["all"], np.array([[0.9, 2.1]]))  # avoid floating errors
    crd.integrals_calc()

    # check that sum agrees -> sqrt of sqsum for uncertainty
    crd_integrals_sum = np.array(crd.integrals_pkg)
    crd_integrals_sum[:, :, 1] = crd_integrals_sum[:, :, 1] ** 2
    crd_integrals_sum = crd_integrals_sum.sum(axis=0)
    crd_integrals_sum[:, 1] = np.sqrt(crd_integrals_sum[:, 1])

    np.testing.assert_almost_equal(crd.integrals, crd_integrals_sum)


def test_calculation_with_filters(crd_data, macro_path):
    """Test that reapplying all filters results in the correct data.

    Ensured that all filters do something on this file!

    :param crd_data: Fixture to return the data path to the CRD files.
    """
    fname = crd_data.joinpath("ti_standard_01.crd")
    crd = CRDFileProcessor(fname)
    crd.spectrum_full()
    crd.def_mcal = np.array([[1.41150472, 45.95262772], [1.84462075, 46.95175879]])
    crd.mass_calibration()

    crd.def_integrals = (
        ["46Ti", "47Ti", "48Ti", "49Ti", "50Ti"],
        np.array(
            [
                [45.81992524, 46.11577284],
                [46.83425987, 47.11601949],
                [47.82041853, 48.10217815],
                [48.83475316, 49.08833681],
                [49.8349998, 50.0604075],
            ]
        ),
    )
    crd.spectrum_part([1, 116000])
    crd.filter_max_ions_per_shot(280)
    crd.filter_max_ions_per_time(170, 10)
    crd.filter_max_ions_per_tof_window(20, np.array([2, 2.4]))
    crd.packages(1000)
    crd.filter_max_ions_per_pkg(500)

    macro_file = macro_path.joinpath("ex_max_ions_per_shot_for_filter_test.py")
    crd.run_macro(macro_file)

    crd.dead_time_correction(7)
    crd.integrals_calc()  # default conditions

    crd2 = CRDFileProcessor(fname)
    crd2.applied_filters = crd.applied_filters
    crd2.calculate_applied_filters()

    np.testing.assert_equal(crd.data, crd2.data)
    np.testing.assert_equal(crd.data_pkg, crd2.data_pkg)
    np.testing.assert_equal(crd.tof, crd2.tof)


def test_calculation_tof_spectrum_never_cut(crd_data):
    """Ensure tof spectrum length stays the same throughout filters"""
    fname = crd_data.joinpath("ti_standard_01.crd")
    crd = CRDFileProcessor(fname)
    interfacer.load_cal_file(crd)
    crd.spectrum_full()
    crd.mass_calibration()

    # cut the range, the set max ions per shot
    crd.spectrum_part([[1, 10000]])
    crd.filter_max_ions_per_shot(10)

    assert crd.data.shape == crd.tof.shape
    assert crd.data.shape == crd.mass.shape


def test_optimize_mcal(crd_data):
    """Optimize the mass calibration of a given spectrum.

    Ensure that it runs through. Assert that values are as before to within 1%.

    :param crd_data: Fixture to return the data path to the CRD files.
    """
    fname = crd_data.joinpath("ti_standard_01.crd")
    mcal_input = np.array([[1.41150472, 45.95262772], [1.84462075, 46.95175879]])
    crd = CRDFileProcessor(fname)
    crd.spectrum_full()
    crd.def_mcal = mcal_input
    crd.mass_calibration()
    crd.optimize_mcal()

    np.testing.assert_allclose(mcal_input, crd.def_mcal, rtol=0.01)



"""CRD reader to handle any kind of header and version (currently v1)."""

from pathlib import Path
import struct
from typing import Tuple
import warnings

import numpy as np

from . import crd_utils


class CRDReader:
    """Read CRD Files and make the data available.

    Example:
        >>> fname = Path("folder/my_crd_file.crd")
        >>> crd_file = CRDReader(fname)
        >>> crd_file.nof_ions
        13281
        >>> crd_file.nof_shots
        5000
    """

    def __init__(self, fname: Path) -> None:
        """Read in a CRD file and make all header arguments available.

        :param fname: Filename

        :raises TypeError: Fname must be a valid path.
        """
        if not isinstance(fname, Path):
            raise TypeError("Filename must be given as a valid Path using pathlib.")
        self.fname = fname

        # header dictionary
        self.header = {}

        # data
        self._ions_per_shot = None
        self._all_tofs = None
        self._ions_to_tof_map = None

        # some quick-available variables
        self._nof_shots = None
        self._nof_ions = None

        # init end of file
        self.eof = False

        # now read the stuff
        self.read_data()

    @property
    def all_data(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get the data.

        :return: 1D array with ions per shot, 1D array with bin in which ions arrived
        """
        return self._ions_per_shot, self._all_tofs

    @property
    def all_tofs(self) -> np.ndarray:
        """Get all time of flight bins.

        :return: 1D array with bin in which ions arrived
        """
        return self._all_tofs

    @property
    def ions_per_shot(self) -> np.ndarray:
        """Get ions per shot array.

        :return: 1D array with ions per shot
        """
        return self._ions_per_shot

    @property
    def ions_to_tof_map(self) -> np.ndarray:
        """Get the index mapper for ions_per_shot -> tof.

        :return: Mapper with indexes where tofs are in all_tofs
        """
        return self._ions_to_tof_map

    @property
    def nof_ions(self) -> int:
        """Get the number of shots.

        :return: Number of shots.
        """
        return self._nof_ions

    @property
    def nof_shots(self) -> int:
        """Get the number of ions.

        :return: Number of ions.
        """
        return self._nof_shots

    # FUNCTIONS #

    def parse_data(self, data: bytes) -> None:
        """Parse the actual data out and put into the appropriate array.

        For this parsing to work, everything has to be just right, i.e., the number
        of shots have to be exactly defined and the data should have the right length.
        If not, this needs to throw a warning and move on to parse in a slower way.

        :param data: Binary string of all the data according to CRD specification.

        :warning: Number of Shots do not agree with the number of shots in the list or
            certain ions are outside the binRange. Fallback to slower reading routine.
        :warning: There is more data in this file than indicated by the number of Shots.
        """
        nof_shots = self.header["nofShots"]
        self._nof_shots = nof_shots

        ions_per_shot = np.zeros(nof_shots, dtype=np.int32)
        # calculate number of ions from filesize
        nof_ions = len(data) // 4 - nof_shots
        self._nof_ions = nof_ions
        all_tofs = np.zeros(nof_ions, dtype=np.int32)

        # loop through the data
        shot_ind = 0  # index in ions_per_shot array
        all_tof_ind = 0  # index in all_tofs array
        bin_ind = 0  # index in binary file
        warning_occured = False  # bool if we had a warning and need to take it slow
        while shot_ind < nof_shots:
            # ions in the given shot
            curr_ions_in_shot = struct.unpack("<I", data[bin_ind : bin_ind + 4])[0]
            ions_per_shot[shot_ind] = curr_ions_in_shot
            # now write out the times
            for _ in range(curr_ions_in_shot):
                try:
                    bin_ind += 4
                    curr_time_bin = struct.unpack("<I", data[bin_ind : bin_ind + 4])[0]
                    all_tofs[all_tof_ind] = curr_time_bin
                    all_tof_ind += 1
                except IndexError:
                    warning_occured = True
                    shot_ind = nof_shots  # to break out of while loop
                    break
            bin_ind += 4
            shot_ind += 1

        if warning_occured or bin_ind != len(data):
            warnings.warn(
                f"This CRD file does not adhere to the specifications and might be "
                f"corrupt. I will try a slow reading routine now in order to get the "
                f"data. Some information: \n"
                f"nof_shots: {self.nof_shots}\n"
                f"nof_ions: {self.nof_ions}\n",
                UserWarning,
            )
            self.parse_data_fallback(data)
            return
        else:
            self._ions_per_shot = ions_per_shot
            self._all_tofs = all_tofs

    def parse_data_fallback(self, data: bytes) -> None:
        """Slow reading routine in case the CRD file is corrupt.

        Here we don't assume anything and just try to read the data into lists and
        append them. Sure, this is going to be slow, but better than no data at all.

        :param data: Array of all the data.
        """
        ions_per_shot = []
        all_tofs = []

        bin_ind = 0  # index inside of binary `data`
        while bin_ind < len(data):
            curr_ions_in_shot = struct.unpack("<I", data[bin_ind : bin_ind + 4])[0]
            ions_per_shot.append(curr_ions_in_shot)
            # now append the times
            for _ in range(curr_ions_in_shot):
                bin_ind += 4
                curr_time_bin = struct.unpack("<I", data[bin_ind : bin_ind + 4])[0]
                all_tofs.append(curr_time_bin)
            bin_ind += 4

        self._nof_shots = len(ions_per_shot)
        self._ions_per_shot = np.array(ions_per_shot)
        self._nof_ions = len(all_tofs)
        self._all_tofs = np.array(all_tofs)

    def read_data(self) -> None:
        """Read in the data and parse out the header.

        The header information will be stored in the header dictionary. All entry
        names are as specified in the CRD format file for version 1.0.

        :raises KeyError: Header is not available.
        :raises OSError: Corrupt data length.
        """
        with open(self.fname, "rb") as f_in:
            # read start of the header
            for (name, size, fmt) in crd_utils.HEADER_START:
                self.header[name] = struct.unpack(fmt, f_in.read(size))[0]

            # get the rest of the header
            crd_version = f"v{self.header['majVer']}p{self.header['minVer']}"
            try:
                hdr_description = crd_utils.CRDHeader[crd_version].value
            except KeyError as exc:
                raise KeyError(
                    f"The header version of this CRD file is {crd_version}, "
                    f"which is not available."
                ) from exc

            for (name, size, fmt) in hdr_description:
                self.header[name] = struct.unpack(fmt, f_in.read(size))[0]

            # now read in the rest of the file
            rest = f_in.read()

        if len(rest) % 4 != 0:
            raise OSError(
                "Data length does not agree with CRD format and seems to be corrupt."
            )

        # check for eof
        if struct.unpack("4s", rest[-4:])[0][:-1] == b"OK!":
            self.eof = True

        # prepare the data
        if self.eof:
            self.parse_data(rest[:-4])
        else:
            self.parse_data(rest)

        # now call the mapper routine
        self._ions_to_tof_map = crd_utils.shot_to_tof_mapper(self._ions_per_shot)



"""Test Excel writer function."""

from pathlib import Path

import numpy as np

from rimseval import CRDFileProcessor
import rimseval.data_io.excel_writer as exw


def test_workup_file_writer_no_integrals(crd_file):
    """Assure nothing is done if no integrals were set."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    # create a file for excel outptut - which won't get generated
    ex_fname = Path("test.xlsx")
    exw.workup_file_writer(crd, ex_fname)

    assert not ex_fname.is_file()


def test_workup_file_run_through(crd_file, tmpdir):
    """Run through excel writer and assure it creates an xlsx file."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()

    # set some random mass cal from 1 to 2
    crd.def_mcal = np.array([[crd.tof.min(), 1.0], [crd.tof.max(), 2.0]])
    crd.mass_calibration()

    # now set the integrals to include everything
    crd.def_integrals = (
        ["Fe54", "Fe56", "FeO"],
        np.array([[0.9, 1.2], [1.2, 2.0], [2.0, 2.1]]),
    )
    crd.integrals_calc()

    # create a file for excel outptut - which won't get generated
    ex_fname = tmpdir.join("test.xlsx")
    exw.workup_file_writer(crd, Path(ex_fname), timestamp=True)
    assert ex_fname.isfile()



"""Process multiple CRD files, open them, handle individually, enable batch runs."""

import gc
from pathlib import Path
from typing import List

import numpy as np
from PyQt6 import QtCore

import rimseval
from rimseval.processor import CRDFileProcessor


class MultiFileProcessor(QtCore.QObject):
    """Class to process multiple CRD files at ones.

    Example:
        >>> file_names = [Path("file1.crd"), Path("file2.crd"), Path("file3.crd")]
        >>> mfp = MultiFileProcessor(file_names)
        >>> mfp.num_of_files
        3
        >>> mfp.peak_fwhm = 0.02  # set full with half max for all files
    """

    signal_processed = QtCore.pyqtSignal(str)  # emits the filename that was processed

    def __init__(self, crd_files: List[Path]):
        """Initialize the multiple file processor.

        :param crd_files: List of pathes to the CRD files.
        """
        super().__init__()
        self._num_of_files = len(crd_files)
        self.crd_files = crd_files

        self._files = None

    # PROPERTIES #

    @property
    def files(self) -> List[CRDFileProcessor]:
        """Return a list of files.

        If the files are not opened, it will open and read them.

        :return: List of CRDFileProcessor instances with files opened
        """
        if self._files is None:
            self.open_files()
        return self._files

    @property
    def num_of_files(self) -> int:
        """Get the number of files that are in the multiprocessor."""
        return self._num_of_files

    @property
    def peak_fwhm(self) -> float:
        """Get / Set FWHM of each peak.

        The getter returns the average, the setter sets the same for all.

        :return: Average peak FWHM in us.
        """
        fwhm = np.zeros(len(self.files))
        for it, file in enumerate(self.files):
            fwhm[it] = file.peak_fwhm
        return np.average(fwhm)

    @peak_fwhm.setter
    def peak_fwhm(self, value: float):
        for file in self.files:
            file.peak_fwhm = value

    # METHODS #

    def apply_to_all(self, id: int, opt_mcal: bool = False, bg_corr: bool = False):
        """Take the configuration for the ID file and apply it to all files.

        :param id: Index where the main CRD file is in the list
        :param opt_mcal: Optimize mass calibration if True (default: False)
        :param bg_corr: Perform background correction?
        """
        crd_main = self.files[id]

        mcal = crd_main.def_mcal
        integrals = crd_main.def_integrals
        backgrounds = crd_main.def_backgrounds
        applied_filters = crd_main.applied_filters

        # run file itself first:
        if crd_main.tof is None:
            crd_main.spectrum_full()
        if crd_main.mass is None:
            crd_main.mass_calibration()
        crd_main.calculate_applied_filters()
        if bg_corr and backgrounds is not None:
            bg_corr = True
        else:
            bg_corr = False
        crd_main.integrals_calc(bg_corr=bg_corr)
        crd_main.integrals_calc_delta()
        self.signal_processed.emit(str(crd_main.fname.name))
        rimseval.interfacer.save_cal_file(crd_main)

        for it, file in enumerate(self.files):
            if it != id:  # skip already done file
                file.spectrum_full()
                if mcal is not None:
                    file.def_mcal = mcal
                    file.mass_calibration()
                    if opt_mcal:
                        file.optimize_mcal()
                if integrals is not None:
                    file.def_integrals = integrals
                if backgrounds is not None:
                    file.def_backgrounds = backgrounds
                if applied_filters is not None:
                    file.applied_filters = applied_filters

                # run the evaluation
                file.calculate_applied_filters()

                # integrals
                file.integrals_calc(bg_corr=bg_corr)
                file.integrals_calc_delta()

                # save calibration
                rimseval.interfacer.save_cal_file(file)

                # emit signal
                self.signal_processed.emit(str(file.fname.name))

    def close_files(self) -> None:
        """Destroys the files and frees the memory."""
        del self._files
        gc.collect()  # garbage collector
        self._files = None

    def close_selected_files(self, ids: List[int], main_id: int = None) -> int:
        """Close selected files.

        Close the ided files and free the memory. If the main_id is given, the program
        will return the new ID of the main file in case it got changed. If the main
        file is gone or no main file was provided, zero is returned.

        :param ids: List of file IDs, i.e., where they are in ``self.files``.
        :param main_id: ID of the main file, an updated ID will be returned if main
            file is present, otherwise zero will be returned.

        :return: New ID of main file if present or given. Otherwise, return zero.
        """
        main_file = None
        if main_id is not None:
            main_file = self._files[main_id]

        ids.sort(reverse=True)
        for id in ids:
            del self._files[id]
        gc.collect()

        if main_id is None:
            return 0
        elif main_file not in self._files:
            return 0
        else:
            return self._files.index(main_file)

    def load_calibrations(self, secondary_cal: Path = None) -> None:
        """Load calibration files for all CRDs.

        This routine checks first if a calibration file with the same name exists.
        If so, it will be loaded. If no primary calibration is present and a secondary
        calibration filename is given, the program will try to load that one if it
        exists. Otherwise, no calibration will be loaded.

        :param secondary_cal: File for secondary calibration. Will only be used if it
            exists.
        """
        for crd in self.files:
            self.load_calibration_single(crd, secondary_cal=secondary_cal)

    @staticmethod
    def load_calibration_single(
        crd: CRDFileProcessor, secondary_cal: Path = None
    ) -> None:
        """Load a single calibration for a CRD file.

        Loads the primary calibration (i.e., the one with the same name but .json
        qualifier) if it exists. Otherwise, if a secondary calibration is given,
        it will try to load that one. If that file does not exist either, nothing
        will be done.

        :param crd: CRD file to load the calibration for.
        :param secondary_cal: Optional, calibration to fall back on if no primary
            calibration is available.
        """
        if (calfile := crd.fname.with_suffix(".json")).is_file():
            pass
        elif secondary_cal is not None and (calfile := secondary_cal).is_file():
            pass
        else:
            return

        rimseval.interfacer.load_cal_file(crd, calfile)

    def open_files(self) -> None:
        """Open the files and store them in the list."""
        files = [CRDFileProcessor(fname) for fname in self.crd_files]
        self._files = files

    def open_additional_files(
        self, fnames: List[Path], read_files=True, secondary_cal=None
    ) -> None:
        """Open additional files to the ones already opened.

        Files will be appended to the list, no sorting.

        :param fnames: List of filenames for the additional files to be opened.
        :param read_files: Read the files after opening?
        :param secondary_cal: Secondary calibration file for loading calibrations.
        """
        for fname in fnames:
            file_to_add = CRDFileProcessor(fname)
            if read_files:
                file_to_add.spectrum_full()
            self.load_calibration_single(file_to_add, secondary_cal=secondary_cal)

            self._files.append(file_to_add)

    def read_files(self) -> None:
        """Run spectrum_full on all CRD files."""
        for file in self.files:
            file.spectrum_full()



"""Functions for fitting using scipy optimize."""

from typing import Union

import numpy as np


def gaussian(
    xdat: Union[float, np.ndarray], coeffs: np.ndarray
) -> Union[float, np.ndarray]:
    """Return the value of a Gaussian for given parameters.

    :param xdat: X value
    :param coeffs: Coefficients, [mu, sigma, height]

    :return: Value of the Gaussian.
    """
    return coeffs[2] * np.exp(-(((xdat - coeffs[0]) / coeffs[1]) ** 2))


def residuals_gaussian(
    coeffs: np.ndarray, ydat: np.ndarray, xdat: np.ndarray
) -> np.ndarray:
    """Calculate residuals and return them.

    :param coeffs: Coefficients for model
    :param ydat: Y data to compare with
    :param xdat: X data for model fit

    :return: Residual of the Gaussian.
    """
    return ydat - gaussian(xdat, coeffs)



"""Example: Re-implementation of maximum ions per shot."""

import numpy as np

from rimseval.processor import CRDFileProcessor


def calc(crd: CRDFileProcessor) -> None:
    """Macro to filter out all shots with more than 3 ions.

    :param crd: CRD file processor that will be passed to the macro.
    """
    # maximum ions per shot to be filtered out
    max_ions = 3

    # create a numpy mask that selects the filters we want to filter
    shots_rejected = np.where(crd.ions_per_shot > max_ions)[0]

    # pass the rejected shots array to the routine that will filter everything
    crd.apply_individual_shots_filter(shots_rejected)



"""Matplotlib Canvas implementation to handle various mouse events."""

from matplotlib.backends.backend_qtagg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from PyQt6 import QtCore, QtWidgets

try:
    import qdarktheme
except ImportError:
    qdarktheme = None

from rimseval.processor import CRDFileProcessor


class PlotSpectrum(QtWidgets.QMainWindow):
    """QMainWindow to plot a ToF or mass spectrum."""

    def __init__(
        self, crd: CRDFileProcessor, logy: bool = True, theme: str = None
    ) -> None:
        """Get a PyQt5 window to define the mass calibration for the given data.

        :param crd: The CRD file processor to work with.
        :param logy: Display the y axis logarithmically? Bottom set to 0.7
        :param theme: Theme, if applicable ("dark" or "light", default None)
        """
        super().__init__()
        self.setWindowTitle("Mass Spectrum")

        self.theme = theme
        if theme is not None and qdarktheme is not None:
            self.setStyleSheet(qdarktheme.load_stylesheet(theme))

        if theme == "dark":
            plt.style.use("dark_background")

        self.crd = crd
        self.logy = logy

        # create a matpotlib canvas using my own canvas
        self.fig = Figure(figsize=(9, 6), dpi=100)
        sc = MplCanvasRightClick(self.fig)
        self.axes = self.fig.add_subplot(111)
        self.sc = sc

        toolbar = MyMplNavigationToolbar(sc, self)
        self.status_bar = QtWidgets.QStatusBar()
        self.setStatusBar(self.status_bar)

        # layouts to populate
        self.bottom_layout = QtWidgets.QHBoxLayout()
        self.right_layout = QtWidgets.QVBoxLayout()

        self.right_layout.addStretch()  # layout must contain something to work...

        # toolbar and canvas
        layout_plot = QtWidgets.QVBoxLayout()
        layout_plot.addWidget(toolbar)
        layout_plot.addWidget(sc)

        # layout top
        layout_top = QtWidgets.QHBoxLayout()
        layout_top.addLayout(layout_plot)
        layout_top.addLayout(self.right_layout)

        # main layout
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(layout_top)

        # add a logy toggle button
        self.button_logy_toggle = QtWidgets.QPushButton("LogY")
        self.button_logy_toggle.setCheckable(True)
        self.button_logy_toggle.setChecked(logy)
        self.button_logy_toggle.clicked.connect(self.logy_toggle)
        self.bottom_layout.addWidget(self.button_logy_toggle)

        # final layout
        layout.addLayout(self.bottom_layout)

        widget = QtWidgets.QWidget()
        widget.setLayout(layout)

        self.setCentralWidget(widget)

    def logy_toggle(self):
        """Toggle logy."""
        self.logy = not self.logy
        self.button_logy_toggle.setChecked(self.logy)

        if self.logy:
            self.axes.set_yscale("log")
            self.axes.set_ylim(bottom=0.7)
        else:
            self.axes.set_yscale("linear")
            self.axes.set_ylim(bottom=0)

        self.sc.draw()

    def plot_tof(self):
        """Plot ToF spectrum."""
        self._plot_data("tof")

    def plot_ms(self):
        """Plot mass spectrum."""
        if self.crd.mass is not None:
            self._plot_data("mass")

    # PRIVATE METHODS #

    def _plot_data(self, case: str) -> None:
        """Plot the data on the canvas.

        :param case: What should we plot on x axis? "tof" for time of flight or
            "mass" for mass spectrum
        """
        if case == "tof":
            xax = self.crd.tof
            xlabel = "Time of Flight (us)"
        else:
            xax = self.crd.mass
            xlabel = "Mass (amu)"

        color = "w" if self.theme == "dark" else "k"

        self.axes.fill_between(xax, self.crd.data, color=color, linewidth=0.3)
        self.axes.set_xlabel(xlabel)
        self.axes.set_ylabel("Counts")
        if self.logy:
            self.axes.set_yscale("log")
            self.axes.set_ylim(bottom=0.7)

        self.sc.draw()


class MplCanvasRightClick(FigureCanvas):
    """MPL Canvas reimplementation to catch right click.

    On right click, emits the coordinates of the position in axes coordinates as
    a signal of two floats (x_position, y_position).
    """

    mouse_right_press_position = QtCore.pyqtSignal(float, float)
    mouse_right_release_position = QtCore.pyqtSignal(float, float)

    def __init__(self, figure: Figure) -> None:
        """Initialize MPL canvas with right click capability.

        :param figure: Matplotlib Figure
        """
        super().__init__(figure)

        self.mpl_connect(
            "button_press_event",
            lambda event: self.emit_mouse_position(event, "pressed"),
        )
        self.mpl_connect(
            "button_release_event",
            lambda event: self.emit_mouse_position(event, "released"),
        )

    def emit_mouse_position(self, event, case):
        """Emit a signal on a right mouse click event.

        Here, bring up a box to ask for the mass, then send it, along with the time
        the mass is at, to the parent class receiver.

        :param event: PyQt event.
        :param case: Which case are we handling? Currently implemented are "pressed",
            "released".
        """
        if event.button == 3:  # right click as an mpl MouseEvent
            if event.xdata is not None and event.ydata is not None:
                if case == "pressed":
                    self.mouse_right_press_position.emit(event.xdata, event.ydata)
                elif case == "released":
                    self.mouse_right_release_position.emit(event.xdata, event.ydata)


class MyMplNavigationToolbar(NavigationToolbar):
    """My own reimplementation of the navigation toolbar.

    Features:
    - untoggle zoom button after zoom is finished.
    """

    def __init__(self, *args, **kwargs):
        """Initialize toolbar."""
        super().__init__(*args, **kwargs)

    def release_zoom(self, event):
        """Run a normal zoom release event and then untoggle button."""
        super().release_zoom(event)
        self.zoom()  # untoggle zoom button



"""PyTest fixtures for func tests in data_io."""

import pytest

import rimseval


@pytest.fixture
def mpa4a_data_ascii():
    """Provides data for the MCS6 TDC, header states [MPA4A] format.

    Provide data in ASCII format, also provide the correct formatting and channel that contains data.

    :return: A tuple with channel, format, and data
    :rtype: (int, DataFormat, list)
    """
    channel = 4
    fmt = rimseval.lst_processor.LST2CRD.ASCIIFormat.ASCII_1A
    data = [
        "000200b95a54",
        "000300b95a54",
        "000400b95a64",
        "000500b95a64",
        "000600b95a54",
        "000700b95a64",
        "000800b95a54",
        "000900b95a64",
        "000a00b95a54",
    ]
    return channel, fmt, data



"""Fixtures for functional tests."""

from pathlib import Path

import pytest
import numpy as np

from rimseval.processor import CRDFileProcessor


@pytest.fixture
def crd_int_delta(crd_file) -> CRDFileProcessor:
    """Initialize a CRDFileProcessor file, set some integrals, and return it.

    The integrals set are fake, no actual calculation takes place. This fixture simply
    serves to test functionality (i.e., delta calculation) after processing of the CRD
    file.

    :return: CRDFileProcessor instance with integrals defined and set.
    """
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    peak_names = ["54Fe", "56Fe", "57Fe", "58Fe", "244Pu", "238Pu", "bg"]
    peak_limits = np.array(
        [
            [53.8, 54.2],
            [55.8, 56.2],
            [56.8, 57.2],
            [57.8, 58.2],
            [243.8, 244.2],
            [237.8, 238.2],
            [238.5, 238.7],
        ]
    )
    crd.def_integrals = peak_names, peak_limits
    crd.integrals = np.array(
        [
            [5.84500000e04, 2.41764348e02],
            [9.17540000e05, 9.57883083e02],
            [2.11900000e04, 1.45567854e02],
            [2.82000000e03, 5.31036722e01],
            [10000, 100],
            [144, 12],
            [34212, 185],
        ]
    )
    crd.integrals_pkg = np.array([crd.integrals, crd.integrals])
    return crd


@pytest.fixture
def data_files_path(request) -> Path:
    """Provides the path to the `data_files` folder.

    :return: Path to the folder
    """
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("data_files").absolute()



"""Configuration file for testing the package with ``nox``."""

import nox

# fixme: add safety to nox
nox.options.sessions = ["lint", "tests"]

package = "rimseval"
locations = "rimseval", "noxfile.py"
python_default = "3.9"
python_suite = ["3.10", "3.9", "3.8"]


@nox.session(python=python_default)
def docs(session):
    """Build the documentation."""
    session.install(".[doc]")
    session.chdir("docs")
    session.run(
        "sphinx-build", "-b", "html", ".", "_build/html/"
    )  # as for readthedocs.io


@nox.session(python=python_default)
def lint(session):
    """Lint project using ``flake8``."""
    args = session.posargs or locations
    session.install(".[dev]")
    session.run("flake8", *args)


@nox.session(python=python_suite)
def tests(session):
    """Test the project using ``pytest``."""
    session.install(".[test]")
    session.run("pytest")


@nox.session(python=python_default)
def safety(session):
    """Safety check for all dependencies."""
    session.install("safety", ".")
    session.run(
        "safety",
        "check",
        "--full-report",
    )



"""Plotting capability for specialty functions."""

import itertools
import sys
from typing import Tuple

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from numba import njit
import numpy as np
from PyQt6 import QtWidgets
from scipy.stats import poisson

try:
    import qdarktheme
except ImportError:
    qdarktheme = None

from rimseval.guis.mpl_canvas import MyMplNavigationToolbar
from rimseval.processor import CRDFileProcessor


# markers to cycle through
MARKERS = ("o", "^", "s", "v", "<", ">", "+", "x", "*", "1", "2", "3", "4")


class PlotFigure(QtWidgets.QMainWindow):
    """QMainWindow to plot a Figure."""

    def __init__(self, logy: bool = False, theme: str = None) -> None:
        """Get a PyQt5 window to define the mass calibration for the given data.

        :param logy: Display the y axis logarithmically? Bottom set to 0.7
        :param theme: Theme, if applicable ("dark" or "light", default None)
        """
        super().__init__()
        self.setWindowTitle("Figure")

        self.theme = theme
        if theme is not None and qdarktheme is not None:
            self.setStyleSheet(qdarktheme.load_stylesheet(theme))

        if theme == "dark":
            plt.style.use("dark_background")
            self.main_color = "w"
        else:
            self.main_color = "tab:blue"

        self.logy = logy

        # create a matpotlib canvas using my own canvas
        self.fig = Figure(figsize=(9, 6), dpi=100)
        self.axes = self.fig.add_subplot(111)
        sc = FigureCanvas(self.fig)
        self.sc = sc

        toolbar = MyMplNavigationToolbar(sc, self)
        self.status_bar = QtWidgets.QStatusBar()
        self.setStatusBar(self.status_bar)

        # layouts to populate
        self.bottom_layout = QtWidgets.QHBoxLayout()

        # toolbar and canvas
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(toolbar)
        layout.addWidget(sc)

        # add a logy toggle button
        self.button_logy_toggle = QtWidgets.QPushButton("LogY")
        self.button_logy_toggle.setCheckable(True)
        self.button_logy_toggle.setChecked(logy)
        self.button_logy_toggle.clicked.connect(self.logy_toggle)
        self.bottom_layout.addWidget(self.button_logy_toggle)

        close_button = QtWidgets.QPushButton("Close")
        close_button.clicked.connect(self.close)
        self.bottom_layout.addStretch()
        self.bottom_layout.addWidget(close_button)

        # final layout
        layout.addLayout(self.bottom_layout)

        widget = QtWidgets.QWidget()
        widget.setLayout(layout)

        self.setCentralWidget(widget)

    def logy_toggle(self):
        """Toggle logy."""
        self.logy = not self.logy
        self.button_logy_toggle.setChecked(self.logy)

        if self.logy:
            self.axes.set_yscale("log")
            self.axes.set_ylim(bottom=0.7)
        else:
            self.axes.set_yscale("linear")
            self.axes.set_ylim(bottom=0)

        self.sc.draw()


class DtIons(PlotFigure):
    """Plot time differences between ions."""

    def __init__(
        self,
        crd: CRDFileProcessor,
        logy: bool = False,
        theme: str = None,
        max_ns: float = None,
    ) -> None:
        """Initialize the class.

        :param crd: CRD file to process.
        :param logy: Plot with logarithmic y axis? Defaults to ``True``
        :param theme: Theme to plot in, defaults to ``None``.
        :param max_ns: Maximum time to plot in ns. If None, plots all.
        """
        super().__init__(logy=logy, theme=theme)

        self.setWindowTitle("Histogram time difference between ions")

        self.crd = crd
        self.max_ns = max_ns

        self.calc_and_draw()

    def calc_and_draw(self) -> None:
        """Calculate the required data and draw it."""
        ion_ranges = self.crd.ions_to_tof_map[np.where(self.crd.ions_per_shot > 1)]
        spacings, frequency = _calculate_bin_differences(self.crd.all_tofs, ion_ranges)

        # turn spacings to ns
        spacings = spacings.astype(float)
        spacings *= self.crd.crd.header["binLength"] / 1000  # bins are in ps

        # plot
        self.axes.plot(spacings, frequency, "-", color=self.main_color)
        self.axes.text(
            0.95,
            0.95,
            f"TDC bin length: {self.crd.crd.header['binLength']}ps",
            horizontalalignment="right",
            verticalalignment="top",
            transform=self.axes.transAxes,
        )

        # labels
        self.axes.set_xlabel("Time between all ions for individual shots (ns)")
        self.axes.set_ylabel("Frequency")
        self.axes.set_title(f"{self.crd.fname.with_suffix('').name}")

        # ax limit
        if self.max_ns is not None:
            self.axes.set_xlim(right=self.max_ns)
        self.axes.set_xlim(left=0)
        self.axes.set_ylim(bottom=0)

        self.sc.draw()


class IntegralsPerPackage(PlotFigure):
    """Plot integrals of all packages versus package number."""

    def __init__(
        self, crd: CRDFileProcessor, logy: bool = False, theme: str = None
    ) -> None:
        """Initialize the class.

        :param crd: CRD file to process.
        :param logy: Plot with logarithmic y-axis? Defaults to ``True``
        :param theme: Theme to plot in, defaults to ``None``.

        :raises OSError: Packages are not defined
        """
        super().__init__(logy=logy, theme=theme)

        self.setWindowTitle("Integrals per package")

        self.crd = crd

        if crd.integrals_pkg is None:
            raise OSError("Integrals for packages are not available.")

        self.calc_and_draw()

    def calc_and_draw(self) -> None:
        """Create the plot for all the defined integrals."""
        int_names = self.crd.def_integrals[0]
        integrals_pkg = self.crd.integrals_pkg

        xdata = np.arange(len(integrals_pkg)) + 1  # start with 1
        counts = integrals_pkg[:, :, 0]
        errors = integrals_pkg[:, :, 1]

        # plot
        marker = itertools.cycle(MARKERS)
        for it in range(counts.shape[1]):  # loop through all defined integrals
            dat = counts[:, it]
            err = errors[:, it]
            self.axes.errorbar(
                xdata, dat, yerr=err, ls="--", label=int_names[it], marker=next(marker)
            )

        # labels
        self.axes.set_xlabel("Package number")
        self.axes.set_ylabel("Counts in integral")
        self.axes.set_title(
            f"Integrals per package - {self.crd.fname.with_suffix('').name}"
        )
        self.axes.legend()

        self.sc.draw()


class IonsPerShot(PlotFigure):
    """Plot histogram for number of ions per shot."""

    def __init__(
        self, crd: CRDFileProcessor, logy: bool = False, theme: str = None
    ) -> None:
        """Initialize the class.

        :param crd: CRD file to process.
        :param logy: Plot with logarithmic y-axis? Defaults to ``True``
        :param theme: Theme to plot in, defaults to ``None``.
        """
        super().__init__(logy=logy, theme=theme)

        self.setWindowTitle("Histogram ions per shot")

        self.crd = crd

        self.calc_and_draw()

    def calc_and_draw(self) -> None:
        """Calculate the histogram and plot it."""
        xdata, hist = _create_histogram(self.crd.ions_per_shot)

        # theoretical prediction
        lambda_poisson = np.sum(self.crd.ions_per_shot) / self.crd.nof_shots
        theoretical_values = poisson.pmf(xdata, lambda_poisson) * np.sum(hist)

        # plot
        self.axes.bar(xdata, hist, width=1, color=self.main_color, label="Data")
        self.axes.step(
            xdata - 0.5,
            theoretical_values,
            "-",
            color="tab:red",
            label="Poisson Distribution",
        )

        # labels
        self.axes.set_xlabel("Number of ions in individual shot")
        self.axes.set_ylabel("Frequency")
        self.axes.set_title(
            f"Histogram number of ions per shot - {self.crd.fname.with_suffix('').name}"
        )
        self.axes.legend()

        self.sc.draw()


def dt_ions(
    crd: CRDFileProcessor, logy: bool = False, theme: str = None, max_ns: float = None
) -> None:
    """Plot ToF difference between ions for shots with 2+ ions.

    :param crd: CRD file to process.
    :param logy: Plot with logarithmic y axis? Defaults to ``True``
    :param theme: Theme to plot in, defaults to ``None``.
    :param max_ns: Maximum time to plot in ns. If None, plots all.
    """
    app = QtWidgets.QApplication(sys.argv)
    window = DtIons(crd, logy=logy, theme=theme, max_ns=max_ns)
    window.show()
    app.exec()


def integrals_packages(
    crd: CRDFileProcessor, logy: bool = False, theme: str = None
) -> None:
    """Plot all the integrals versus package number for data split into packages.

    :param crd: CRD file to process.
    :param logy: Plot with logarithmic y axis? Defaults to ``True``
    :param theme: Theme to plot in, defaults to ``None``.
    """
    app = QtWidgets.QApplication(sys.argv)
    window = IntegralsPerPackage(crd, logy=logy, theme=theme)
    window.show()
    app.exec()


def nof_ions_per_shot(
    crd: CRDFileProcessor, logy: bool = False, theme: str = None
) -> None:
    """Plot a histogram of the number of shots in a given crd file.

    The histogram is compared with the theoretical curve based on poisson statistics.

    :param crd: CRD file to process.
    :param logy: Plot with logarithmic y axis? Defaults to ``True``
    :param theme: Theme to plot in, defaults to ``None``.
    """
    app = QtWidgets.QApplication(sys.argv)
    window = IonsPerShot(crd, logy=logy, theme=theme)
    window.show()
    app.exec()


@njit
def _create_histogram(data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Sort the data into a histogram. Bins are equal to one integer."""
    xdata = np.arange(np.max(data) + 1)
    hist = np.zeros_like(xdata)
    for it in data:
        hist[int(it)] += 1
    return xdata, hist


@njit
def _calculate_bin_differences(
    all_tofs: np.ndarray, ion_ranges: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Sort through bins and write the bin spacings back.

    :param all_tofs: All tofs array with info on arrival bins
    :param ion_ranges: Range of ions to consider.

    :return: Spacings between ions in bins, frequency of spacing occurance
    """
    # calculate number of spacings -> must be a gaussian sum of numbers
    nof_spacings = 0
    for rng in ion_ranges:
        nof_ions = rng[1] - rng[0]
        nof_spacings += nof_ions * (nof_ions - 1) / 2

    ind = 0
    spacings = np.zeros(int(nof_spacings), dtype=np.int32)
    for rng in ion_ranges:
        ions = all_tofs[rng[0] : rng[1]]
        for it in range(len(ions) - 1):
            diffs = np.abs(ions[it + 1 :] - ions[it])
            spacings[ind : ind + len(diffs)] = diffs
            ind += len(diffs)

    # now create the histogram
    min_diff = np.min(spacings)
    max_diff = np.max(spacings)

    frequency = np.zeros(max_diff - min_diff + 1, dtype=np.int32)
    for sp in spacings:
        frequency[sp - min_diff] += 1

    return np.arange(min_diff, max_diff + 1), frequency



"""PyTest fixtures for unit tests."""

from pathlib import Path
import pytest


@pytest.fixture
def legacy_files_path(request):
    """Provides the path to the `legacy_files` folder.

    :return: Path to the folder
    :rtype: Path
    """
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("../legacy_files").absolute()



"""Export data to files."""

from pathlib import Path

from numba import njit
import numpy as np

from rimseval.processor import CRDFileProcessor


@njit
def _bin_array_avg(arr: np.ndarray, bins: int) -> np.ndarray:  # pragma: nocover
    """Take a numpy array and bins it by averaging the range.

    End of the array, if it doesn't fit, will be thrown away.

    :param arr: Array of data.
    :param bins: Number of entries to bin.

    :return: Binned array.
    """
    ret_array = np.zeros(len(arr) // bins)
    for it in range(len(ret_array)):
        ret_array[it] = np.average(arr[it * bins : it * bins + bins])
    return ret_array


@njit
def _bin_array_sum(arr: np.ndarray, bin: int) -> np.ndarray:  # pragma: nocover
    """Take a numpy array and bins it by summing the range.

    End of the array, if it doesn't fit, will be thrown away.

    :param arr: Array of data.
    :param bin: Number of entries to bin.

    :return: Binned array.
    """
    ret_array = np.zeros(len(arr) // bin)
    for it in range(len(ret_array)):
        ret_array[it] = np.sum(arr[it * bin : it * bin + bin])
    return ret_array


def tof_spectrum(crd: CRDFileProcessor, fname: Path, bins: int = 1) -> None:
    """Export time of flight spectra to csv file.

    :param crd: CRD file.
    :param fname: File name to export to.
    :param bins: How many data to bin.
    """
    tof_binned = _bin_array_avg(crd.tof, bins)
    data_binned = _bin_array_sum(crd.data, bins)

    header = "Time of Flight (us),Counts"
    data_to_write = np.stack([tof_binned, data_binned], axis=1)

    fname = fname.with_suffix(".csv").absolute()

    np.savetxt(fname, data_to_write, delimiter=",", header=header)


def mass_spectrum(crd: CRDFileProcessor, fname: Path, bins: int = 1) -> None:
    """Export time of flight and mass spectra to csv file.

    :param crd: CRD file.
    :param fname: File name to export to.
    :param bins: How many data to bin.
    """
    tof_binned = _bin_array_avg(crd.tof, bins)
    mass_binned = _bin_array_avg(crd.mass, bins)
    data_binned = _bin_array_sum(crd.data, bins)

    header = "Mass (amu),Time of Flight (us),Counts"
    data_to_write = np.stack([mass_binned, tof_binned, data_binned], axis=1)

    fname = fname.with_suffix(".csv").absolute()

    np.savetxt(fname, data_to_write, delimiter=",", header=header)



"""Function tests for interfacer."""

from pathlib import Path

import pytest
import numpy as np

from rimseval import interfacer


def test_read_lion_eval_calfile(mocker, crd_proc_mock):
    """Set crd file mass cal, integrals, and bg_corr from LIONEval cal file.

    Mocking getting a file since the LIONEvalCal is actually tested in its own unit
    tests with proper files. Here we just fake some data.
    """
    mcal_exp = np.array([[1, 1], [2, 2]])  # expected and returned
    integrals_ret = [["46Ti", 46.0, 0.2, 0.3], ["47Ti", 47.1, 0.3, 0.3]]
    integrals_exp = (["46Ti", "47Ti"], np.array([[45.8, 46.3], [46.8, 47.4]]))
    backgrounds_ret = [["46Ti", 46.0, 44.2, 44.4, 46.4, 46.6]]
    backgrounds_exp = [["46Ti", "46Ti"], np.array([[44.2, 44.4], [46.4, 46.6]])]

    cal_mock = mocker.MagicMock()
    cal_mock_prop_mcal = mocker.PropertyMock(return_value=mcal_exp)
    cal_mock_prop_integrals = mocker.PropertyMock(return_value=integrals_ret)
    cal_mock_prop_backgrounds = mocker.PropertyMock(return_value=backgrounds_ret)
    mocker.patch("rimseval.interfacer.LIONEvalCal", cal_mock)
    type(cal_mock()).mass_cal = cal_mock_prop_mcal
    type(cal_mock()).integrals = cal_mock_prop_integrals
    type(cal_mock()).bg_corr = cal_mock_prop_backgrounds

    interfacer.read_lion_eval_calfile(crd=crd_proc_mock, fname=Path("."))

    np.testing.assert_equal(crd_proc_mock.def_mcal, mcal_exp)
    integrals_rec = crd_proc_mock.def_integrals
    backgrounds_rec = crd_proc_mock.def_backgrounds
    assert integrals_rec[0] == integrals_exp[0]  # names
    np.testing.assert_almost_equal(integrals_rec[1], integrals_exp[1])
    assert backgrounds_rec[0] == backgrounds_exp[0]
    np.testing.assert_almost_equal(backgrounds_rec[1], backgrounds_exp[1])


def test_read_lion_eval_calfile_read_error(data_files_path, crd_proc_mock):
    """Raise IOError if the calibration file cannot be read successfully."""
    cal_file = data_files_path.joinpath("faulty_cal_file.json")

    err_exp = f"Cannot open the calibration file {cal_file.name}. JSON decode error."

    with pytest.raises(IOError) as err:
        interfacer.load_cal_file(crd_proc_mock, cal_file)

    msg = err.value.args[0]
    assert msg == err_exp



"""Tests for the plots utility functions that are in GUIs."""

import numpy as np

from rimseval.guis import plots


def test_create_histogram():
    """Create a histogram from a data array."""
    data = np.array([1, 2, 1, 1, 1, 2, 5, 4, 3, 4, 3, 1, 1, 0, 0, 0])
    xdata_exp = np.array([0, 1, 2, 3, 4, 5])
    hist_exp = np.array([3, 6, 2, 2, 2, 1])

    xdata_rec, hist_rec = plots._create_histogram(data)
    np.testing.assert_equal(xdata_exp, xdata_rec)
    np.testing.assert_equal(hist_exp, hist_rec)


def test_calculate_bin_differences():
    """Create a spacing vector with spaces in beetween ions."""
    # test with only part of array being used
    all_tofs = np.arange(20)
    ion_ranges = np.array([[2, 5], [7, 10]])  # ranges with > 1 ion per shot
    spacing_exp = np.array([1, 2])
    frequency_exp = np.array([4, 2])

    spacing_rec, frequency_rec = plots._calculate_bin_differences(all_tofs, ion_ranges)
    np.testing.assert_equal(spacing_exp, spacing_rec)
    np.testing.assert_equal(frequency_exp, frequency_rec)

    # test with zeros in array
    all_tofs = np.array([5, 10, 11, 9, 10, 15, 20, 25])
    ion_ranges = np.array([[0, 3], [3, 8]])
    spacing_exp = np.arange(1, 16 + 1)
    frequency_exp = np.array([2, 0, 0, 0, 4, 2, 0, 0, 0, 2, 1, 0, 0, 0, 1, 1])

    spacing_rec, frequency_rec = plots._calculate_bin_differences(all_tofs, ion_ranges)
    np.testing.assert_equal(spacing_exp, spacing_rec)
    np.testing.assert_equal(frequency_exp, frequency_rec)


def test_calculate_bin_differences_negative():
    """Ensure that the absolute is taken between arrival times."""
    all_tofs = np.array([2, 1])
    ion_ranges = np.array([[0, 2]])
    spacing_exp = np.array([1])
    frequency_exp = np.array([1])

    spacing_rec, frequency_rec = plots._calculate_bin_differences(all_tofs, ion_ranges)
    np.testing.assert_equal(spacing_exp, spacing_rec)
    np.testing.assert_equal(frequency_exp, frequency_rec)



"""Unit tests for the CRD reader, making use of the `crd_file` fixture."""

from pathlib import Path

import pytest
import numpy as np

from rimseval.data_io.crd_reader import CRDReader
from ...utils import assert_crd_equal


# TEST PROPERTIES #


def test_crd_reader_wrong_filetype():
    """Raise TypeError if file is not given as Path instance."""
    fname = "some_file.crd"
    with pytest.raises(TypeError) as err:
        CRDReader(fname)
    err_msg = err.value.args[0]
    assert err_msg == "Filename must be given as a valid Path using pathlib."


def test_crd_reader_header(crd_file):
    """Assert that the header of a crd file is read correctly."""
    hdr, _, _, fname = crd_file
    crd = CRDReader(Path(fname))
    assert crd.header == hdr


def test_crd_reader_all_data(crd_file):
    """Return all the data from the CRD file."""
    _, ions_per_shot, all_tofs, fname = crd_file
    crd = CRDReader(Path(fname))
    ret_ions_per_shot, ret_arrival_bins = crd.all_data
    np.testing.assert_equal(ions_per_shot, ret_ions_per_shot)
    np.testing.assert_equal(all_tofs, ret_arrival_bins)


def test_crd_all_tofs(crd_file):
    """Return all tof arrival bins."""
    _, _, all_tofs, fname = crd_file
    crd = CRDReader(Path(fname))
    np.testing.assert_equal(crd.all_tofs, all_tofs)


def test_crd_ions_per_shot(crd_file):
    """Return ions per shot array."""
    _, ions_per_shot, _, fname = crd_file
    crd = CRDReader(Path(fname))
    np.testing.assert_equal(crd.ions_per_shot, ions_per_shot)


def test_crd_nof_ions(crd_file):
    """Return number of ions."""
    _, ions_per_shot, _, fname = crd_file
    crd = CRDReader(Path(fname))
    assert crd.nof_ions == ions_per_shot.sum()


def test_crd_nof_shots(crd_file):
    """Return number of shots."""
    hdr, _, _, fname = crd_file
    crd = CRDReader(Path(fname))
    assert crd.nof_shots == hdr["nofShots"]


# ERROR CHECKING #


def test_invalid_data_length(crd_data):
    """Raise an OSError if an incorrect data length was encountered."""
    crd_file = crd_data.joinpath("err_invalid_data_length.crd")
    with pytest.raises(OSError) as err:
        _ = CRDReader(crd_file)

    msg = err.value.args[0]
    assert "Data length does not agree" in msg


def test_invalid_header(crd_data):
    """Raise a KeyError if an incorrect header was encountered."""
    hdr_version = "v0p1"
    crd_file = crd_data.joinpath("err_invalid_header.crd")
    with pytest.raises(KeyError) as err:
        _ = CRDReader(crd_file)

    msg = err.value.args[0]
    assert hdr_version in msg


def test_inavlid_num_of_ions_less(crd_data):
    """Parse the file properly, even if less ions in file than expected."""
    crd_file = crd_data.joinpath("err_less_ions_than_hdr.crd")

    with pytest.warns(UserWarning, match="I will try a slow reading routine now"):
        _ = CRDReader(crd_file)


def test_inavlid_num_of_ions_more(crd_data):
    """Parse the file properly, even if more ions in file than expected."""
    crd_file = crd_data.joinpath("err_less_ions_than_hdr.crd")

    with pytest.warns(UserWarning, match="I will try a slow reading routine now"):
        _ = CRDReader(crd_file)


def test_no_eof(crd_data, crd_file):
    """Parse file even without an EoF."""
    _, _, _, fname = crd_file
    crd_correct = CRDReader(Path(fname))
    crd_faulty = CRDReader(Path(crd_data.joinpath("err_no_eof.crd")))

    assert_crd_equal(crd_correct, crd_faulty)



"""Test list file utilities."""

import pathlib
import pytest
import numpy as np

import rimseval.data_io.lst_utils as utl


def test_ascii_to_ndarray_ascii_1a_no_tag(init_lst_proc):
    """Convert ASCII_1A, data to numpy array, w/o tag."""
    data_list = [
        "0001000e7474",
        "0002000e7474",
        "0002000e7473",  # wrong channel
        "0003000e7474",
        "000000000000",  # zero
    ]
    fmt = init_lst_proc.ASCIIFormat.ASC_1A
    channel = 4
    expected_return = np.array([[1, 59207], [2, 59207], [3, 59207]], dtype=np.uint32)

    ret_data, ret_tag, _ = utl.ascii_to_ndarray(data_list, fmt, channel)
    np.testing.assert_equal(ret_data, expected_return)
    assert ret_tag is None


def test_ascii_to_ndarray_ascii_1a_tag(init_lst_proc):
    """Convert ASCII_1A, data to numpy array, with tag."""
    data_list = [
        "0001000e7474",
        "0002000e7474",
        "0002000e7473",  # tag
        "0003000e7474",
        "000000000000",  # zero
    ]
    fmt = init_lst_proc.ASCIIFormat.ASC_1A
    channel = 4
    tag = 3
    expected_data = np.array([[1, 59207], [2, 59207], [3, 59207]], dtype=np.uint32)
    expected_tag = np.array([2], dtype=np.uint32)

    ret_data, ret_tag, _ = utl.ascii_to_ndarray(data_list, fmt, channel, tag)

    np.testing.assert_equal(ret_data, expected_data)
    np.testing.assert_equal(ret_tag, expected_tag)


def test_ascii_to_ndarray_ascii_1a_no_tag_other_channel(init_lst_proc):
    """Convert ASCII_1A, data to numpy array, w/o tag."""
    data_list = [
        "0001000e7474",
        "0002000e7474",
        "0002000e7473",  # wrong channel
        "0003000e7474",
        "000000000000",  # zero
    ]
    fmt = init_lst_proc.ASCIIFormat.ASC_1A
    other_channels_exp = [4]
    channel = 3

    _, _, other_channels_ret = utl.ascii_to_ndarray(data_list, fmt, channel)
    assert other_channels_ret == other_channels_exp


def test_get_sweep_time_ascii():
    """Transfer binary number to base 10 int based on boundaries."""
    bin_str = "1000101101"
    sweep_boundaries = (0, 5)
    time_boundaries = (5, 10)
    sweep_exp = int(bin_str[sweep_boundaries[0] : sweep_boundaries[1]], 2)
    time_exp = int(bin_str[time_boundaries[0] : time_boundaries[1]], 2)

    sweep_ret, time_ret = utl.get_sweep_time_ascii(
        bin_str, sweep_boundaries, time_boundaries
    )
    assert sweep_ret == sweep_exp
    assert time_ret == time_exp


def test_separate_signal_with_tag():
    """Separate a signal into tagged and untagged data."""
    signal_all = np.array([[1, 9207], [2, 5207], [3, 59207]], dtype=np.uint32)
    tag_location = np.array([2], dtype=np.uint32)

    untagged_exp = np.array([[1, 9207], [3, 59207]], dtype=np.uint32)
    tagged_exp = np.array([[2, 5207]], dtype=np.uint32)

    untagged_rec, tagged_rec = utl.separate_signal_with_tag(signal_all, tag_location)

    np.testing.assert_equal(untagged_rec, untagged_exp)
    np.testing.assert_equal(tagged_rec, tagged_exp)


def test_transfer_lst_to_crd_data():
    max_sweep = 1023
    data_in = np.array(
        [
            [1, 500],
            [2, 265],
            [1, 600],
            [5, 700],
            [500, 5000],
            [550, 2],
            [501, 55],
            [800, 50],
            [1023, 200],
            [2, 100],
            [1023, 13],
            [5, 5000],
        ],
        dtype=np.uint32,
    )
    ion_range = 1000

    # expected shots array
    shots_array_exp = np.zeros(1023 + 5, dtype=np.uint32)
    shots_array_exp[0] = 2
    shots_array_exp[1] = 1
    shots_array_exp[4] = 1
    shots_array_exp[500] = 1
    shots_array_exp[549] = 1
    shots_array_exp[799] = 1
    shots_array_exp[1022] = 2
    shots_array_exp[1024] = 1

    # expected ions array
    ions_array_exp = np.array(
        [500, 600, 265, 700, 55, 2, 50, 200, 13, 100], dtype=np.uint32
    )

    shots_array_ret, ions_array_ret, out_of_range = utl.transfer_lst_to_crd_data(
        data_in, max_sweep, ion_range
    )

    np.testing.assert_equal(shots_array_ret, shots_array_exp)
    np.testing.assert_equal(ions_array_ret, ions_array_exp)
    assert out_of_range  # some ions are above ion range



"""Routine to pre-process list files.

This routine turns list files into crd files. The crd file format is as specified in
the `docs` folder and adheres currently to v1.0 of the format.
"""

from . import crd_utils
from . import export
from . import lst_utils
from .crd_reader import CRDReader
from .lst_to_crd import LST2CRD

__all__ = ["export", "crd_utils", "lst_utils", "CRDReader", "LST2CRD"]



"""Class to turn a list file to a CRD file."""

from datetime import datetime
from enum import Enum
from pathlib import Path
import struct
import warnings

import numpy as np

from . import crd_utils
from . import lst_utils


class LST2CRD:
    """Convert list files to CRD files.

    Example:
        >>> from pathlib import Path
        >>> from rimseval.data_io import LST2CRD
        >>> file = Path("path/to/file.lst")
        >>> lst = LST2CRD(file_name=file, channel_data=1, tag_data=None)
        >>> lst.read_list_file()
        >>> lst.write_crd()
    """

    class BinWidthTDC(Enum):
        """Bin width defined by instrument.

        These are only FASTComTec instruments. The name of each entry is equal to
        the identifier that can be found in a datafile (lst).
        The entries of the enum are as following:

        - binwidth in ps
        """

        MPA4A = 100
        MCS8A = 80

    class ASCIIFormat(Enum):
        """Available formats for this routine that are already implemented.

        Various formats that are implemented when dealing with ASCII data.
        The value is composed of a tuple of 2 entries.

         - 0: entry: width of the binary number (binary_width)
         - 1: Tuple of tuples with start, stop on where to read
           0: sweep - 1: time - 2: channel
        """

        ASC_1A = (48, ((0, 16), (16, 44), (45, 48)))
        ASC_9 = (64, ((1, 21), (21, 59), (60, 64)))

    class DATFormat(Enum):
        """Available formats (time_patch) for binary data.

        Various binary data formats are incorporated. Value is compmosed of 2 entries:

          - 0: Data length in bytes
          - 1: Encoding of the binary value to read with struct.unpack()
          - 2: Tuple, Where in the decoded list are: 0: sweep - 1: time - 2: channel
        """

        DAT_9 = (8, "<")

    def __init__(
        self,
        file_name: Path = None,
        channel_data: int = None,
        channel_tag: int = None,
    ) -> None:
        """Initialize the LST2CRD class.

        :param file_name: File name and path of file to be read.
        :param channel_data: Number of channel the data are in.
        :param channel_tag: Number of channel the tag is in, None for no tag.
        """
        # set the default values
        self._channel_data = channel_data
        self._channel_tag = channel_tag
        self._file_name = file_name

        # initialize values for future use
        self._binary_file = False  # is this a binary file?
        self._file_info = {}  # dictionary with parsed header info
        self._data_format = None  # format of the data. auto set on reading
        self._data_signal = None  # data for signal (total)
        self._tags = None  # array for tagged shots
        self._other_channels = None  # List for other channels that have counts

    # PROPERTIES #

    @property
    def channel_data(self) -> int:
        """Get / set the channel number of the data.

        :return: Channel number of data

        :raises TypeError: Channel number is not an integer.
        """  # noqa: DAR402
        return self._channel_data

    @channel_data.setter
    def channel_data(self, newval: int) -> None:
        if not isinstance(newval, int):
            raise TypeError("Channel number must be given as an integer.")
        self._channel_data = newval

    @property
    def channel_tag(self) -> int:
        """Get / set the channel number of the tag.

        :return: Channel number of tag

        :raises TypeError: Channel number is not an integer.
        """  # noqa: DAR402
        return self._channel_tag

    @channel_tag.setter
    def channel_tag(self, newval):
        if not isinstance(newval, int):
            raise TypeError("Channel number must be given as an integer.")
        self._channel_tag = newval

    @property
    def data_format(self) -> ASCIIFormat:
        """Select the data format to use to convert the LST file to CRD.

        :return: The currently chosen data format.

        :raises TypeError: Data format is not a DataFormat enum.
        """  # noqa: DAR402
        return self._data_format

    @data_format.setter
    def data_format(self, newval: ASCIIFormat) -> None:
        if not isinstance(newval, self.ASCIIFormat):
            raise TypeError(
                f"Your data format {newval} is not a valid type. "
                f"You must choose an object from the `DataFormat` instance."
            )
        self._data_format = newval

    @property
    def file_name(self) -> Path:
        """Get / set the file name for the file to be read / written.

        :return: The path and file name to the selected object.

        :raises TypeError: Path is not a `pathlib.Path` object.
        """  # noqa: DAR402
        return self._file_name

    @file_name.setter
    def file_name(self, newval: Path) -> None:
        if not isinstance(newval, Path):
            raise TypeError(
                f"Path must be a `pathlib.Path` object but is a {type(newval)}."
            )
        self._file_name = newval

    # METHODS #

    def read_list_file(self) -> None:
        """Read a list file specified in `self.file_name`.

        This routine sets the following parameters of the class:

        - self._file_data
        - self._tag_data (if a tag was selected)

        This routine sets the following information parameters in self._file_info:

        - "bin_width": Sets the binwidth in ps, depending on the instrument
        - "calfact": Calibration factor, to scale range to bins
        - "data_type": Sets the data type, 'ascii' for ASCII or 'dat' for binary, str
        - "shot_range": shot range
        - "timestamp": Time and date of file recording
        - "time_patch": Data format, as reported by Fastcomtec as time_patch. as str

        :raises ValueError: File name not provided.
        :raises ValueError: Channel for data not provided.
        :raises OSError: The Data Format is not available / could not be found in file.
        :raises NotImplementedError: The current data format is not (yet) implemented.
        """
        if self.file_name is None:
            raise ValueError("Please set a file name.")
        if self.channel_data is None:
            raise ValueError("Please set a number for the data channel.")

        # read in the file
        try:
            with self.file_name.open() as f:
                content = f.read().splitlines()
        except UnicodeDecodeError:  # we might have a binary file:
            with self.file_name.open("rb") as f:
                content = f.read().splitlines()
                self._binary_file = True

        # find the data and save it into a data_ascii list
        fnd_str = b"[DATA]" if self._binary_file else "[DATA]"
        index_start_data = content.index(fnd_str) + 1
        header = content[:index_start_data]
        if self._binary_file:
            header = [it.decode("utf-8") for it in header]
        data_ascii = content[index_start_data:]

        # set the bin width - in ps
        bin_width = None
        for it in self.BinWidthTDC:
            if it.name.lower() in header[0].lower():
                bin_width = it.value
                break
        if bin_width is None:
            raise NotImplementedError(
                f"The current data format cannot be identified. "
                f"The datafile header starts with: {header[0]}. "
                f"Available instruments are the following: "
                f"{[it.name for it in self.BinWidthTDC]}"
            )
        else:
            self._file_info["bin_width"] = bin_width

        # find calfact - in ns
        calfact = None
        for head in header:
            if head[0:8] == "calfact=":
                calfact = float(head.split("=")[1])
                self._file_info["calfact"] = calfact
                break

        # find the range
        for head in header:
            if head[0:5] == "range":
                ion_range = int(head.split("=")[1])
                mult_fact = calfact / (bin_width / 1000)  # get range in bin_width
                self._file_info["ion_range"] = int(ion_range * mult_fact)
                break

        # find the data type, ascii or binary
        data_type = None
        for head in header:
            if head[0:6] == "mpafmt":
                data_type = head.split("=")[1]
                self._file_info["data_type"] = data_type
                break
        if data_type is None:
            raise OSError("Could not find a data type in the list file!")

        # find the time patch
        for head in header:
            if head[0:10] == "time_patch":
                time_patch = head.split("=")[1]
                self._file_info["time_patch"] = time_patch

        # Find timestamp
        for head in header:
            tmp_date_str = "cmline0="
            if head[0 : len(tmp_date_str)] == tmp_date_str:
                datetime_str = head.replace(tmp_date_str, "").split()
                date_tmp = datetime_str[0].split("/")  # month, day, year
                time_tmp = datetime_str[1].split(":")  # h, min, sec
                self._file_info["timestamp"] = datetime(
                    year=int(date_tmp[2]),
                    month=int(date_tmp[0]),
                    day=int(date_tmp[1]),
                    hour=int(time_tmp[0]),
                    minute=int(time_tmp[1]),
                    second=int(float(time_tmp[2])),
                )
                break

        # find the data format or raise an error
        self.set_data_format()

        # currently, only ascii data are supported. error checking in set_data_format()
        data_sig, tags, other_channels = lst_utils.ascii_to_ndarray(
            data_ascii, self._data_format, self.channel_data, self.channel_tag
        )

        self._data_signal = data_sig
        self._tags = tags
        self._other_channels = other_channels

        # set number of ions
        self._file_info["no_ions"] = len(data_sig)

    def write_crd(self) -> None:
        """Write CRD file(s) from the data that are in the class.

        .. note:: A file must have been read first. Also, this routine doesn't actually
            write the crd file itself, but it handles the tags, etc., and then
            sources the actual writing task out.

        :raises ValueError: No data has been read in.
        :raises OSError: Data is empty.
        """
        if self._data_signal is None:
            raise ValueError("No data has been read in yet.")

        if self._data_signal.shape[0] == 0:
            raise OSError(
                f"There are no counts present in this file. Please double "
                f"check that you are using the correct channel for the signal. "
                f"The file seems to have counts in channels {self._other_channels}."
            )

        # calculate the maximum number of sweeps that can be recorded
        max_sweeps = pow(
            2, self.data_format.value[1][0][1] - self.data_format.value[1][0][0]
        )

        ions_out_of_range_warning = False
        if self._channel_tag is not None:  # we have tagged data
            # prepare the data
            untagged_data, tagged_data = lst_utils.separate_signal_with_tag(
                self._data_signal, self._tags
            )
            (
                untagged_shots,
                untagged_ions,
                untagged_ions_out_of_range,
            ) = lst_utils.transfer_lst_to_crd_data(
                untagged_data, max_sweeps, self._file_info["ion_range"]
            )
            (
                tagged_shots,
                tagged_ions,
                tagged_ions_out_of_range,
            ) = lst_utils.transfer_lst_to_crd_data(
                tagged_data, max_sweeps, self._file_info["ion_range"]
            )

            # raise warning later?
            if untagged_ions_out_of_range or tagged_ions_out_of_range:
                ions_out_of_range_warning = True

            # write crd
            fname_untagged = self.file_name.with_suffix(".untagged.crd")
            self._write_crd(fname_untagged, untagged_shots, untagged_ions)
            fname_tagged = self.file_name.with_suffix(".tagged.crd")
            self._write_crd(fname_tagged, tagged_shots, tagged_ions)

        else:  # untagged
            # prepare the data
            (
                data_shots,
                data_ions,
                ions_out_of_range,
            ) = lst_utils.transfer_lst_to_crd_data(
                self._data_signal, max_sweeps, self._file_info["ion_range"]
            )

            # raise warning later?
            if ions_out_of_range:
                ions_out_of_range_warning = True

            # write crd
            fname = self.file_name.with_suffix(".crd")
            self._write_crd(fname, data_shots, data_ions)

        if ions_out_of_range_warning:
            warnings.warn(
                "The lst file contained ions that were outside the allowed "
                "range. These ions were not written to the crd file.",
                UserWarning,
            )

    def set_data_format(self):
        """Set the data format according to what is saved in file_info dictionary.

        The "data_type" and "time_patch" values must be present in the dictionary.
        Writes the information to itself, to the `_data_format` variable.

        :raises NotImplementedError: Binary data are currently not supported.
        :raises ValueError: Needs to be binary or ASCII data.
        """
        data_type = self._file_info["data_type"]
        time_patch = self._file_info["time_patch"]
        fmt_str = f"{data_type.upper()}_{time_patch.upper()}"
        if data_type.lower() == "asc":
            fmt = self.ASCIIFormat[fmt_str]
        elif data_type.lower() == "dat":
            raise NotImplementedError("Binary data is currently not supported.")
        else:
            raise ValueError(
                f"The data type {fmt_str} seems to be neither binary nor ASCII."
            )
        self._data_format = fmt

    def _write_crd(
        self, fname: Path, data_shots: np.ndarray, data_ions: np.ndarray
    ) -> None:
        """Write an actual CRD file as defined.

        Defaults of this writing are populated from the default dictionary in crd_utils.

        :param fname: File name to write to
        :param data_shots: Prepared array with all shots included.
        :param data_ions: Prepared array with all ions included.
        """
        default = crd_utils.CURRENT_DEFAULTS

        # prepare date and time
        dt = self._file_info["timestamp"]
        dt_fmt = (
            f"{dt.year:04}:{dt.month:02}:{dt.day:02} "
            f"{dt.hour:02}:{dt.minute:02}:{dt.second:02}"
        )

        # get the new bin range
        bin_start = data_ions.min()
        bin_end = data_ions.max()

        with open(fname, "wb") as fout:
            # header
            fout.write(default["fileID"])
            fout.write(struct.pack("20s", bytes(dt_fmt, "utf-8")))
            fout.write(default["minVer"])
            fout.write(default["majVer"])
            fout.write(default["sizeOfHeaders"])
            fout.write(default["shotPattern"])
            fout.write(default["tofFormat"])
            fout.write(default["polarity"])
            fout.write(struct.pack("<I", self._file_info["bin_width"]))
            fout.write(struct.pack("<I", bin_start))
            fout.write(struct.pack("<I", bin_end))
            fout.write(default["xDim"])
            fout.write(default["yDim"])
            fout.write(default["shotsPerPixel"])
            fout.write(default["pixelPerScan"])
            fout.write(default["nOfScans"])
            fout.write(struct.pack("<I", len(data_shots)))  # number of shots
            fout.write(default["deltaT"])

            # write the data
            ion_cnt = 0
            for shot in data_shots:
                fout.write(struct.pack("<I", shot))
                for _ in range(shot):
                    fout.write(struct.pack("<I", data_ions[ion_cnt]))
                    ion_cnt += 1

            # EoF
            fout.write(default["eof"])
